import rdkit.Chem as Chem
import rdkit.Chem.AllChem as AllChem
from rdkit.Chem.rdchem import ChiralType, BondType, BondDir, BondStereo
from rdchiral.utils import vprint, PLEVEL

BondDirOpposite = {AllChem.BondDir.ENDUPRIGHT: AllChem.BondDir.ENDDOWNRIGHT,
                   AllChem.BondDir.ENDDOWNRIGHT: AllChem.BondDir.ENDUPRIGHT,
                   AllChem.BondDir.NONE: AllChem.BondDir.NONE}
BondDirLabel    = {AllChem.BondDir.ENDUPRIGHT: '\\',
                   AllChem.BondDir.ENDDOWNRIGHT: '/'}

def bond_dirs_by_mapnum(mol):
    bond_dirs_by_mapnum = {}
    for b in mol.GetBonds():
        i = None; j = None
        if b.GetBeginAtom().GetAtomMapNum():
            i = b.GetBeginAtom().GetAtomMapNum()
        if b.GetEndAtom().GetAtomMapNum():
            j = b.GetEndAtom().GetAtomMapNum()
        if i is None or j is None or b.GetBondDir() == BondDir.NONE:
            continue
        bond_dirs_by_mapnum[(i, j)] = b.GetBondDir()
        bond_dirs_by_mapnum[(j, i)] = BondDirOpposite[b.GetBondDir()]
    return bond_dirs_by_mapnum

def enumerate_possible_cistrans_defs(template_r, \
        labeling_func=lambda a: a.GetAtomMapNum()):
    '''
    This function is meant to take a reactant template and fully enumerate
    all the ways in which different double-bonds can have their cis/trans
    chirality specified (based on labeling_func). This is necessary because
    double-bond chirality cannot be specified using cis/trans (global properties)
    but must be done using ENDUPRIGHT and ENDDOWNRIGHT for the attached single
    bonds (local properties). Now, the next issue is that on each side of the
    double bond, only one of the single bond directions must be specified, and
    that direction can be using either atom order. e.g.,

    A1         B1
       \      /
         C = C
       /      \
    A2         B2

    Can be specified by:
    A1-C is an ENDDOWNRIGHT, C-B1 is an ENDUPRIGHT
    A1-C is an ENDDOWNRIGHT, C-B2 is an ENDDOWNRIGHT
    A1-C is an ENDDOWNRIGHT, B1-C is an ENDDOWNRIGHT
    A1-C is an ENDDOWNRIGHT, B2-C is an ENDUPRIGHT
    ...and twelve more definitions using different A1/A2 specs.

    ALSO - we can think about horizontally reflecting this bond entirely,
    which gets us even more definitions.

    So, the point of this function is to fully enumerate *all* of the ways
    in which chirality could have been specified. That way, we can take a
    reactant atom and check if its chirality is within the list of acceptable
    definitions to determine if a match was made.

    Gross.

    The way we do this is by first defining the *local* chirality of a double
    bond, which weights side chains based purely on the unique mapnum numbering.
    Once we have a local cis/trans definition for a double bond, we can enumerate
    the sixteen possible ways that a reactant could match it.

    '''

    required_bond_defs = {}
    required_bond_defs_coreatoms = set()

    if PLEVEL >= 10: print('Looking at initializing template frag')
    for b in template_r.GetBonds():
        if b.GetBondType() != BondType.DOUBLE:
            continue

        # Define begin and end atoms of the double bond
        ba = b.GetBeginAtom()
        bb = b.GetEndAtom()

        # Now check if it is even possible to specify
        if ba.GetDegree() == 1 or bb.GetDegree() == 1:
            continue

        ba_label = labeling_func(ba)
        bb_label = labeling_func(bb)
            
        if PLEVEL >= 10: print('Found a double bond with potential cis/trans (based on degree)')
        if PLEVEL >= 10: print('{} {} {}'.format(ba_label,
                               b.GetSmarts(),
                               bb_label))

        # Save core atoms so we know that cis/trans was POSSIBLE to specify
        required_bond_defs_coreatoms.add((ba_label, bb_label))
        required_bond_defs_coreatoms.add((bb_label, ba_label))
            
        # Define heaviest mapnum neighbor for each atom, excluding the other side of the double bond
        ba_neighbor_labels = [labeling_func(a) for a in ba.GetNeighbors()]
        ba_neighbor_labels.remove(bb_label) # remove other side of =
        ba_neighbor_labels_max = max(ba_neighbor_labels)
        bb_neighbor_labels = [labeling_func(a) for a in bb.GetNeighbors()]
        bb_neighbor_labels.remove(ba_label) # remove other side of =
        bb_neighbor_labels_max = max(bb_neighbor_labels)

        # The direction of the bond being observed might need to be flipped,
        # based on
        #     (a) if it is the heaviest atom on this side, and 
        #     (b) if the begin/end atoms for the directional bond are 
        #         in the wrong order (i.e., if the double-bonded atom 
        #         is the begin atom)
        front_spec = None
        back_spec = None
        for bab in ba.GetBonds():
            if bab.GetBondDir() != BondDir.NONE:
                if labeling_func(bab.GetBeginAtom()) == ba_label:
                    # Bond is in wrong order - flip
                    if labeling_func(bab.GetEndAtom()) != ba_neighbor_labels_max:
                        # Defined atom is not the heaviest - flip
                        front_spec = bab.GetBondDir()
                        break
                    front_spec = BondDirOpposite[bab.GetBondDir()]
                    break
                if labeling_func(bab.GetBeginAtom()) != ba_neighbor_labels_max:
                    # Defined atom is not heaviest
                    front_spec = BondDirOpposite[bab.GetBondDir()]
                    break
                front_spec = bab.GetBondDir()
                break
        if front_spec is None:
            if PLEVEL >= 10: print('Chirality not specified at front end of the bond!')
        else:
            if PLEVEL >= 10: print('Front specification: {}'.format(front_spec))
            
            for bbb in bb.GetBonds():
                if bbb.GetBondDir() != BondDir.NONE:
                    # For the "back" specification, the double-bonded atom *should* be the BeginAtom
                    if labeling_func(bbb.GetEndAtom()) == bb_label:
                        # Bond is in wrong order - flip
                        if labeling_func(bbb.GetBeginAtom()) != bb_neighbor_labels_max:
                            # Defined atom is not the heaviest - flip
                            back_spec = bbb.GetBondDir()
                            break
                        back_spec = BondDirOpposite[bbb.GetBondDir()]
                        break
                    if labeling_func(bbb.GetEndAtom()) != bb_neighbor_labels_max:
                        # Defined atom is not heaviest - flip
                        back_spec = BondDirOpposite[bbb.GetBondDir()]
                        break
                    back_spec = bbb.GetBondDir()
                    break
        if back_spec is None:
            if PLEVEL >= 10: print('Chirality not specified at back end of the bond!')
        else:
            if PLEVEL >= 10: print('Back specification: {}'.format(back_spec))

        # Is this an overall unspecified bond? Put it in the dictionary anyway, 
        # so there is something to match
        if front_spec is None or back_spec is None:
            # Create a definition over this bond so that reactant MUST be unspecified, too
            for start_atom in ba_neighbor_labels:
                for end_atom in bb_neighbor_labels:
                    required_bond_defs[(start_atom, ba_label, bb_label, end_atom)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(ba_label, start_atom, bb_label, end_atom)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(start_atom, ba_label, end_atom, bb_label)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(ba_label, start_atom, end_atom, bb_label)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(bb_label, end_atom, start_atom, ba_label)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(end_atom, bb_label, start_atom, ba_label)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(bb_label, end_atom, ba_label, start_atom)] = (BondDir.NONE, BondDir.NONE)
                    required_bond_defs[(end_atom, bb_label, ba_label, start_atom)] = (BondDir.NONE, BondDir.NONE)
            continue
        
        if front_spec == back_spec:
            if PLEVEL >= 10: print('-> locally TRANS')
            b.SetProp('localChirality', 'trans')
        else:
            if PLEVEL >= 10: print('--> locally CIS')
            b.SetProp('localChirality', 'cis')

        possible_defs = {}
        for start_atom in ba_neighbor_labels:
            for end_atom in bb_neighbor_labels:
                needs_inversion = (start_atom != ba_neighbor_labels_max) != \
                    (end_atom != bb_neighbor_labels_max)
                for start_atom_dir in [BondDir.ENDUPRIGHT, BondDir.ENDDOWNRIGHT]:
                    # When locally trans, BondDir of start shold be same as end, 
                    # unless we need inversion
                    if (front_spec != back_spec) != needs_inversion: 
                        # locally cis and does not need inversion (True, False)
                        # or locally trans and does need inversion (False, True)
                        end_atom_dir = BondDirOpposite[start_atom_dir]
                    else:
                        # locally cis and does need inversion (True, True)
                        # or locally trans and does not need inversion (False, False)
                        end_atom_dir = start_atom_dir

                    # Enumerate all combinations of atom orders...
                    possible_defs[(start_atom, ba_label, bb_label, end_atom)] = (start_atom_dir, end_atom_dir)
                    possible_defs[(ba_label, start_atom, bb_label, end_atom)] = (BondDirOpposite[start_atom_dir], end_atom_dir)
                    possible_defs[(start_atom, ba_label, end_atom, bb_label)] = (start_atom_dir, BondDirOpposite[end_atom_dir])
                    possible_defs[(ba_label, start_atom, end_atom, bb_label)] = (BondDirOpposite[start_atom_dir], BondDirOpposite[end_atom_dir])

                    possible_defs[(bb_label, end_atom, start_atom, ba_label)] = (end_atom_dir, start_atom_dir)
                    possible_defs[(bb_label, end_atom, ba_label, start_atom)] = (end_atom_dir, BondDirOpposite[start_atom_dir])
                    possible_defs[(end_atom, bb_label, start_atom, ba_label)] = (BondDirOpposite[end_atom_dir], start_atom_dir)
                    possible_defs[(end_atom, bb_label, ba_label, start_atom)] = (BondDirOpposite[end_atom_dir], BondDirOpposite[start_atom_dir])

        # Save to the definition of this bond (in either direction)
        required_bond_defs.update(possible_defs)
        
    if PLEVEL >= 10: print('All bond specs for this template:' )
    if PLEVEL >= 10: print(str([(k, v) for (k, v) in required_bond_defs.items()]))
    return required_bond_defs, required_bond_defs_coreatoms

def get_atoms_across_double_bonds(mol, labeling_func=lambda a:a.GetAtomMapNum()):
    '''
    This function takes a molecule and returns a list of cis/trans specifications
    according to the following:

    (mapnums, dirs)

    where atoms = (a1, a2, a3, a4) and dirs = (d1, d2)
    and (a1, a2) defines the ENDUPRIGHT/ENDDOWNRIGHT direction of the "front"
    of the bond using d1, and (a3, a4) defines the direction of the "back" of 
    the bond using d2.

    This is used to initialize reactants with a SINGLE definition constraining
    the chirality. Templates have their chirality fully enumerated, so we can
    match this specific definition to the full set of possible definitions
    when determining if a match should be made.

    NOTE: the atom mapnums are returned. This is so we can later use them
    to get the old_mapno property from the corresponding product atom, which is
    an outcome-specific assignment

    We also include implicit chirality here based on ring membership, but keep
    track of that separately
    '''
    atoms_across_double_bonds = []
    atomrings = None

    for b in mol.GetBonds():
        if b.GetBondType() != BondType.DOUBLE:
            continue

        # Define begin and end atoms of the double bond
        ba = b.GetBeginAtom()
        bb = b.GetEndAtom()

        # Now check if it is even possible to specify
        if ba.GetDegree() == 1 or bb.GetDegree() == 1:
            continue

        ba_label = labeling_func(ba)
        bb_label = labeling_func(bb)
            
        if PLEVEL >= 5: print('Found a double bond with potential cis/trans (based on degree)')
        if PLEVEL >= 5: print('{} {} {}'.format(ba_label,
                               b.GetSmarts(),
                               bb_label))
        
        # Try to specify front and back direction separately
        front_mapnums = None 
        front_dir = None 
        back_mapnums = None 
        back_dir = None
        is_implicit = False 
        bab = None; bbb = None;
        for bab in (z for z in ba.GetBonds() if z.GetBondType() != BondType.DOUBLE):
            if bab.GetBondDir() != BondDir.NONE:
                front_mapnums = (labeling_func(bab.GetBeginAtom()), labeling_func(bab.GetEndAtom()))
                front_dir = bab.GetBondDir()
                break 
        for bbb in (z for z in bb.GetBonds() if z.GetBondType() != BondType.DOUBLE):
            if bbb.GetBondDir() != BondDir.NONE:
                back_mapnums = (labeling_func(bbb.GetBeginAtom()), labeling_func(bbb.GetEndAtom()))
                back_dir = bbb.GetBondDir()
                break 

        # If impossible to spec, just continue
        if (bab is None or bbb is None):
            continue

        # Did we actually get a specification out?
        if (front_dir is None or back_dir is None):

            if b.IsInRing(): 
                # Implicit cis! Now to figure out right definitions...
                if atomrings is None:
                    atomrings = mol.GetRingInfo().AtomRings() # tuple of tuples of atomIdx
                for atomring in atomrings:
                    if ba.GetIdx() in atomring and bb.GetIdx() in atomring:
                        front_mapnums = (labeling_func(bab.GetOtherAtom(ba)), ba_label)
                        back_mapnums = (bb_label, labeling_func(bbb.GetOtherAtom(bb)))
                        if (bab.GetOtherAtomIdx(ba.GetIdx()) in atomring) != \
                                (bbb.GetOtherAtomIdx(bb.GetIdx()) in atomring):
                            # one of these atoms are in the ring, one is outside -> trans
                            if PLEVEL >= 10: print('Implicit trans found')
                            front_dir = BondDir.ENDUPRIGHT
                            back_dir = BondDir.ENDUPRIGHT
                        else:
                            if PLEVEL >= 10: print('Implicit cis found')
                            front_dir = BondDir.ENDUPRIGHT 
                            back_dir = BondDir.ENDDOWNRIGHT
                        is_implicit = True 
                        break

            else:
                # Okay no, actually unspecified
                # Specify direction as BondDir.NONE using whatever bab and bbb were at the end fo the loop
                # note: this is why we use "for bab in ___generator___", so that we know the current
                #       value of bab and bbb correspond to a single bond we can def. by
                front_mapnums = (labeling_func(bab.GetBeginAtom()), labeling_func(bab.GetEndAtom()))
                front_dir = BondDir.NONE
                back_mapnums = (labeling_func(bbb.GetBeginAtom()), labeling_func(bbb.GetEndAtom()))
                back_dir = BondDir.NONE

        # Save this (a1, a2, a3, a4) -> (d1, d2) spec
        atoms_across_double_bonds.append(
            (
                front_mapnums + back_mapnums,
                (front_dir, back_dir),
                is_implicit,
            )
        )

    return atoms_across_double_bonds

def restore_bond_stereo_to_sp2_atom(a, bond_dirs_by_mapnum):
    '''Copy over single-bond directions (ENDUPRIGHT, ENDDOWNRIGHT) to 
    the single bonds attached to some double-bonded atom, a

    a - atom with a double bond
    bond_dirs_by_mapnum - dictionary of (begin_mapnum, end_mapnum): bond_dir
        that defines if a bond should be ENDUPRIGHT or ENDDOWNRIGHT. The reverse
        key is also included with the reverse bond direction. If the source
        molecule did not have a specified chirality at this double bond, then
        the mapnum tuples will be missing from the dict

    In some cases, like C=C/O>>C=C/Br, we should assume that stereochem was
    preserved, even though mapnums won't match. There might be some reactions
    where the chirality is inverted (like C=C/O >> C=C\Br), but let's not
    worry about those for now...

    Returns True if a bond direction was copied'''

    for bond_to_spec in a.GetBonds():
        if (bond_to_spec.GetOtherAtom(a).GetAtomMapNum(), a.GetAtomMapNum()) in bond_dirs_by_mapnum:
            bond_to_spec.SetBondDir(
                bond_dirs_by_mapnum[
                    (bond_to_spec.GetBeginAtom().GetAtomMapNum(),
                     bond_to_spec.GetEndAtom().GetAtomMapNum())
                ]
            )
            if PLEVEL >= 2: print('Tried to copy bond direction b/w {} and {}'.format(
                bond_to_spec.GetBeginAtom().GetAtomMapNum(),
                bond_to_spec.GetEndAtom().GetAtomMapNum()
            ))
            return True
    
    # Weird case, like C=C/O >> C=C/Br
    if PLEVEL >= 2: print('Bond stereo could not be restored to sp2 atom, missing the branch that was used to define before...')
    
    if a.GetDegree() == 2:
        # Either the branch used to define was replaced with H (deg 3 -> deg 2)
        # or the branch used to define was reacted (deg 2 -> deg 2)
        for bond_to_spec in a.GetBonds():
            if bond_to_spec.GetBondType() == BondType.DOUBLE:
                continue
            if not bond_to_spec.GetOtherAtom(a).HasProp('old_mapno'): 
                # new atom, deg2->deg2, assume direction preserved
                if PLEVEL >= 5: print('Only single-bond attachment to atom {} is new, try to reproduce chirality'.format(a.GetAtomMapNum()))
                needs_inversion = False 
            else:
                # old atom, just was not used in chirality definition - set opposite
                if PLEVEL >= 5: print('Only single-bond attachment to atom {} is old, try to reproduce chirality'.format(a.GetAtomMapNum()))
                needs_inversion = True

            for (i, j), bond_dir in bond_dirs_by_mapnum.items():
                if bond_dir != BondDir.NONE:
                    if i == bond_to_spec.GetBeginAtom().GetAtomMapNum():
                        if needs_inversion:
                            bond_to_spec.SetBondDir(BondDirOpposite[bond_dir])
                        else:
                            bond_to_spec.SetBondDir(bond_dir)
                        return True 

    if a.GetDegree() == 3:
        # If we lost the branch defining stereochem, it must have been replaced
        for bond_to_spec in a.GetBonds():
            if bond_to_spec.GetBondType() == BondType.DOUBLE:
                continue
            oa = bond_to_spec.GetOtherAtom(a)
            if oa.HasProp('old_mapno') or oa.HasProp('react_atom_idx'):
                # looking at an old atom, which should have opposite direction as removed atom
                needs_inversion = True 
            else:
                # looking at a new atom, assume same as removed atom
                needs_inversion = False 

            for (i, j), bond_dir in bond_dirs_by_mapnum.items():
                if bond_dir != BondDir.NONE:
                    if i == bond_to_spec.GetBeginAtom().GetAtomMapNum():
                        if needs_inversion:
                            bond_to_spec.SetBondDir(BondDirOpposite[bond_dir])
                        else:
                            bond_to_spec.SetBondDir(bond_dir)
                        return True 

    return False


from retro_star.api import RSPlanner
from time import time
import argparse
import os
import networkx as nx
import json
import pandas as pd

parser = argparse.ArgumentParser()
parser.add_argument('product',              type=str)
parser.add_argument('output',               type=str)
parser.add_argument('-b', '--blocks',       type=str, default='data/building_block.csv')
parser.add_argument('-i', '--iterations',   type=int, default=100)
parser.add_argument('-e', '--exp_topk',     type=int, default=10)
parser.add_argument('-k', '--route_topk',   type=int, default=10)
parser.add_argument('-s', '--beam_size',    type=int, default=10)
parser.add_argument('-m', '--model_type',   type=str, default='ensemble', choices=['ensemble','retroformer','g2s','retriever_only'])
parser.add_argument('-r', '--retrieval',    type=str, default='true', choices=['true', 'false'])
parser.add_argument('-pr', '--path_retrieval',    type=str, default='true', choices=['true', 'false'])
parser.add_argument('-d', '--retrieval_db', type=str, default='data/train_canonicalized.txt')
parser.add_argument('-pd', '--path_retrieval_db', type=str, default='data/pathways.pickle')
parser.add_argument('-c', '--device',       type=str, default='cuda', choices=['cuda', 'cpu'])
args = parser.parse_args()

t_start = time()

planner = RSPlanner(
    cuda=args.device=='cuda',
    iterations=args.iterations,
    expansion_topk=args.exp_topk,
    route_topk=args.route_topk,
    beam_size=args.beam_size,
    model_type=args.model_type,
    retrieval=args.retrieval=='true',
    retrieval_db=args.retrieval_db,
    path_retrieval=args.path_retrieval=='true',
    path_retrieval_db=args.path_retrieval_db,
    starting_molecules=args.blocks
)

result = planner.plan(args.product)


def line_to_graph(line,token):
    rs = line.split('|')
    edges = []
    for r in rs:
        es = r.split('>')
        if token in es[2]:
            edges.extend([es])
        else:
            ps = es[2].split('.')
            edges.extend([(es[0],es[1],p)for p in ps])
    edges = pd.DataFrame(edges,columns=['from','weight','to'])
    g = nx.from_pandas_edgelist(edges, 'from', 'to', ['weight'], create_using = nx.DiGraph())
    return g

def graph_to_dict(graph):
    nodes = [{'id': node} for node in graph.nodes()]
    edges = [{'source': source, 'target': target, 'weight': weight} for source, target, weight in graph.edges(data='weight')]
    
    return {'nodes': nodes, 'edges': edges}

def save_to_json(pathways, output_name, token='keggpath'):
    graphs = [line_to_graph(p,token) for p in pathways]
    graph_dicts = {f'g{i+1}': graph_to_dict(graph) for i, graph in enumerate(graphs)}
    with open(output_name, 'w') as json_file:
        json.dump(graph_dicts, json_file, indent=4)

output_name = args.output

if result is None:
    with open(output_name, 'w') as json_file:
        os.system(f'touch {output_name}')
    print(None)
else:
    save_to_json(result, output_name)
    for i, route in enumerate(result):
        print(f'{i} {route}')

print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



import numpy as np
from typing import Dict, List, Tuple
from g2s.utils.data_utils import get_graph_features_from_smi, load_vocab, tokenize_smiles, G2SDataset_single


def get_token_ids(tokens: list, vocab: Dict[str, int], max_len: int) -> Tuple[List, int]:
    token_ids = []
    token_ids.extend([vocab[token] for token in tokens])
    token_ids = token_ids[:max_len-1]
    token_ids.append(vocab["_EOS"])

    lens = len(token_ids)
    while len(token_ids) < max_len:
        token_ids.append(vocab["_PAD"])

    return token_ids, lens


def get_seq_features_from_line_single(smi, vocab, max_src_len=1024):
    src_tokens = smi.strip().split()
    if not src_tokens:
        src_tokens = ["C", "C"]             # hardcode to ignore
    src_token_ids, src_lens = get_token_ids(src_tokens, vocab, max_len=max_src_len)
    src_token_ids = np.array(src_token_ids, dtype=np.int32)
    return src_token_ids, src_lens


def preprocess_smi(smi, vocab):
    smi = tokenize_smiles(''.join(smi.strip().split()))
    src_token_ids, src_lens = get_seq_features_from_line_single(smi, vocab)

    src_token_ids = np.stack([src_token_ids], axis=0)
    src_lens = np.array([src_lens], dtype=np.int32)

    a_scopes, a_scopes_lens, b_scopes, b_scopes_lens, a_features, a_features_lens, \
        b_features, b_features_lens, a_graphs, b_graphs = get_graph_features_from_smi((0, "".join(smi.split()), False))

    a_scopes = np.concatenate([a_scopes], axis=0)
    b_scopes = np.concatenate([b_scopes], axis=0)
    a_features = np.concatenate([a_features], axis=0)
    b_features = np.concatenate([b_features], axis=0)
    a_graphs = np.concatenate([a_graphs], axis=0)
    b_graphs = np.concatenate([b_graphs], axis=0)

    a_scopes_lens = np.array([a_scopes_lens], dtype=np.int32)
    b_scopes_lens = np.array([b_scopes_lens], dtype=np.int32)
    a_features_lens = np.array([a_features_lens], dtype=np.int32)
    b_features_lens = np.array([b_features_lens], dtype=np.int32)

    feat = {'src_token_ids': src_token_ids, 'src_lens': src_lens, 'a_scopes': a_scopes, 'b_scopes': b_scopes,
            'a_features': a_features, 'b_features': b_features, 'a_graphs': a_graphs, 'b_graphs': b_graphs,
            'a_scopes_lens': a_scopes_lens, 'b_scopes_lens': b_scopes_lens,
            'a_features_lens': a_features_lens, 'b_features_lens': b_features_lens}
    return feat


def build_batch(smi, vocab):
    feat = preprocess_smi(smi, vocab)
    dataset = G2SDataset_single(feat, vocab)
    return dataset[0]


if __name__ == "__main__":
    vocab = load_vocab('data/vocab_smiles.txt')
    batch = build_batch('C C ( = O ) N / C = C ( / C C ( = O ) O ) C ( = O ) O', vocab)
    import pdb; pdb.set_trace()



import torch
import torch.nn as nn
import torch.nn.functional as F
import logging


class ValueMLP(nn.Module):
    def __init__(self, n_layers, fp_dim, latent_dim, dropout_rate, device):
        super(ValueMLP, self).__init__()
        self.n_layers = n_layers
        self.fp_dim = fp_dim
        self.latent_dim = latent_dim
        self.dropout_rate = dropout_rate
        self.device = device

        # logging.info('Initializing value model: latent_dim=%d' % self.latent_dim)

        layers = []
        layers.append(nn.Linear(fp_dim, latent_dim))
        # layers.append(nn.BatchNorm1d(latent_dim,
        #                              track_running_stats=False))
        layers.append(nn.ReLU())
        layers.append(nn.Dropout(self.dropout_rate))
        for _ in range(self.n_layers - 1):
            layers.append(nn.Linear(latent_dim, latent_dim))
            # layers.append(nn.BatchNorm1d(latent_dim,
            #                              track_running_stats=False))
            layers.append(nn.ReLU())
            layers.append(nn.Dropout(self.dropout_rate))
        layers.append(nn.Linear(latent_dim, 1))

        self.layers = nn.Sequential(*layers)

    def forward(self, fps):
        x = fps
        x = self.layers(x)
        x = torch.log(1 + torch.exp(x))

        return x



from __future__ import print_function
from rdkit.Chem.rdchem import ChiralType, BondType, BondDir

from rdchiral.utils import vprint, parity4, PLEVEL

def template_atom_could_have_been_tetra(a, strip_if_spec=False, cache=True):
    '''
    Could this atom have been a tetrahedral center?
    If yes, template atom is considered achiral and will not match a chiral rct
    If no, the tempalte atom is auxilliary and we should not use it to remove
    a matched reaction. For example, a fully-generalized terminal [C:1] 
    '''

    if a.HasProp('tetra_possible'):
        return a.GetBoolProp('tetra_possible')
    if a.GetDegree() < 3 or (a.GetDegree() == 3 and 'H' not in a.GetSmarts()):
        if cache:
            a.SetBoolProp('tetra_possible', False)
        if strip_if_spec: # Clear chiral tag in case improperly set
            a.SetChiralTag(ChiralType.CHI_UNSPECIFIED)
        return False 
    if cache:
        a.SetBoolProp('tetra_possible', True)
    return True 



def copy_chirality(a_src, a_new):

    # Not possible to be a tetrahedral center anymore?
    if a_new.GetDegree() < 3:
        return 
    if a_new.GetDegree() == 3 and \
            any(b.GetBondType() != BondType.SINGLE for b in a_new.GetBonds()):
        return

    if PLEVEL >= 3: print('For mapnum {}, copying src {} chirality tag to new'.format(
        a_src.GetAtomMapNum(), a_src.GetChiralTag()))
    a_new.SetChiralTag(a_src.GetChiralTag())
    
    if atom_chirality_matches(a_src, a_new) == -1:
        if PLEVEL >= 3: print('For mapnum {}, inverting chirality'.format(a_new.GetAtomMapNum()))
        a_new.InvertChirality()

def atom_chirality_matches(a_tmp, a_mol):
    '''
    Checks for consistency in chirality between a template atom and a molecule atom.

    Also checks to see if chirality needs to be inverted in copy_chirality

    Returns +1 if it is a match and there is no need for inversion (or ambiguous)
    Returns -1 if it is a match but they are the opposite
    Returns 0 if an explicit NOT match
    Returns 2 if ambiguous or achiral-achiral
    '''
    if a_mol.GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
        if a_tmp.GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
            if PLEVEL >= 3: print('atom {} is achiral & achiral -> match'.format(a_mol.GetAtomMapNum()))
            return 2 # achiral template, achiral molecule -> match
        # What if the template was chiral, but the reactant isn't just due to symmetry?
        if not a_mol.HasProp('_ChiralityPossible'):
            # It's okay to make a match, as long as the product is achiral (even
            # though the product template will try to impose chirality)
            if PLEVEL >= 3: print('atom {} is specified in template, but cant possibly be chiral in mol'.format(a_mol.GetAtomMapNum()))
            return 2

        # Discussion: figure out if we want this behavior - should a chiral template
        # be applied to an achiral molecule? For the retro case, if we have
        # a retro reaction that requires a specific stereochem, return False;
        # however, there will be many cases where the reaction would probably work
        if PLEVEL >= 3: print('atom {} is achiral in mol, but specified in template'.format(a_mol.GetAtomMapNum()))
        return 0
    if a_tmp.GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
        if PLEVEL >= 3: print('Reactant {} atom chiral, rtemplate achiral...'.format(a_tmp.GetAtomMapNum()))
        if template_atom_could_have_been_tetra(a_tmp):
            if PLEVEL >= 3: print('...and that atom could have had its chirality specified! no_match')
            return 0
        if PLEVEL >= 3: print('...but the rtemplate atom could not have had chirality specified, match anyway')
        return 2

    mapnums_tmp = [a.GetAtomMapNum() for a in a_tmp.GetNeighbors()]    
    mapnums_mol = [a.GetAtomMapNum() for a in a_mol.GetNeighbors()]

    # When there are fewer than 3 heavy neighbors, chirality is ambiguous...
    if len(mapnums_tmp) < 3 or len(mapnums_mol) < 3:
        return 2

    # Degree of 3 -> remaining atom is a hydrogen, add to list
    if len(mapnums_tmp) < 4:
        mapnums_tmp.append(-1) # H
    if len(mapnums_mol) < 4:
        mapnums_mol.append(-1) # H

    try:
        if PLEVEL >= 10: print(str(mapnums_tmp))
        if PLEVEL >= 10: print(str(mapnums_mol))
        if PLEVEL >= 10: print(str(a_tmp.GetChiralTag()))
        if PLEVEL >= 10: print(str(a_mol.GetChiralTag()))
        only_in_src = [i for i in mapnums_tmp if i not in mapnums_mol][::-1] # reverse for popping
        only_in_mol = [i for i in mapnums_mol if i not in mapnums_tmp]
        if len(only_in_src) <= 1 and len(only_in_mol) <= 1:
            tmp_parity = parity4(mapnums_tmp)
            mol_parity = parity4([i if i in mapnums_tmp else only_in_src.pop() for i in mapnums_mol])
            if PLEVEL >= 10: print(str(tmp_parity))
            if PLEVEL >= 10: print(str(mol_parity))
            parity_matches = tmp_parity == mol_parity
            tag_matches = a_tmp.GetChiralTag() == a_mol.GetChiralTag()
            chirality_matches = parity_matches == tag_matches
            if PLEVEL >= 2: print('mapnum {} chiral match? {}'.format(a_tmp.GetAtomMapNum(), chirality_matches))
            return 1 if chirality_matches else -1
        else:
            if PLEVEL >= 2: print('mapnum {} chiral match? Based on mapnum lists, ambiguous -> True'.format(a_tmp.GetAtomMapNum()))
            return 2 # ambiguous case, just return for now

    except IndexError as e:
        print(a_tmp.GetPropsAsDict())
        print(a_mol.GetPropsAsDict())
        print(a_tmp.GetChiralTag())
        print(a_mol.GetChiralTag())
        print(str(e))
        print(str(mapnums_tmp))
        print(str(mapnums_mol))
        raise KeyError('Pop from empty set - this should not happen!')


import numpy as np
import logging


class MolNode:
    def __init__(self, mol, init_value, parent=None, is_known=False,
                 zero_known_value=True):
        self.mol = mol
        self.value = init_value
        self.succ_value = np.inf    # total cost for existing solution
        self.parent = parent

        self.id = -1
        if self.parent is None:
            self.depth = 0
        else:
            self.depth = self.parent.depth

        self.is_known = is_known
        self.children = []
        self.succ = is_known
        self.open = True    # before expansion: True, after expansion: False
        if is_known:
            self.open = False
            if zero_known_value:
                self.value = 0
            self.succ_value = self.value

        if parent is not None:
            parent.children.append(self)

    def v_self(self):
        """
        :return: V_self(self | subtree)
        """
        return self.value

    def v_target(self):
        """
        :return: V_target(self | whole tree)
        """
        if self.parent is None:
            return self.value
        else:
            return self.parent.v_target()

    def init_values(self, no_child=False):
        assert self.open and (no_child or self.children)

        new_value = np.inf
        self.succ = False
        for reaction in self.children:
            new_value = np.min((new_value, reaction.v_self()))
            self.succ |= reaction.succ

        v_delta = new_value - self.value
        self.value = new_value

        if self.succ:
            for reaction in self.children:
                self.succ_value = np.min((self.succ_value,
                                          reaction.succ_value))

        self.open = False

        return v_delta

    def backup(self, succ):
        assert not self.is_known

        new_value = np.inf
        for reaction in self.children:
            new_value = np.min((new_value, reaction.v_self()))
        new_succ = self.succ | succ
        updated = (self.value != new_value) or (self.succ != new_succ)

        new_succ_value = np.inf
        if new_succ:
            for reaction in self.children:
                new_succ_value = np.min((new_succ_value, reaction.succ_value))
            updated = updated or (self.succ_value != new_succ_value)

        v_delta = new_value - self.value
        self.value = new_value
        self.succ = new_succ
        self.succ_value = new_succ_value

        if updated and self.parent:
            return self.parent.backup(v_delta, from_mol=self.mol)

    def serialize(self):
        text = '%d | %s' % (self.id, self.mol)
        return text

    def get_ancestors(self):
        if self.parent is None:
            return {self.mol}

        ancestors = self.parent.parent.get_ancestors()
        ancestors.add(self.mol)
        return ancestors


import os
import sys
root = os.path.abspath('.')
if root not in sys.path:
    sys.path.insert(0, root)

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

from time import time
import torch
import numpy as np
from g2s.models.graph2seq_series_rel import Graph2SeqSeriesRel
from g2s.utils.data_utils import load_vocab
from g2s.utils.preprocess import build_batch

from rdkit import RDLogger, Chem
RDLogger.DisableLog("rdApp.*")


def filter_invalid(generations, scores):
    filtered_generations, filtered_scores = [], []
    for generation, score in zip(generations, scores):
        mol = Chem.MolFromSmiles(generation)
        if mol is None:
            continue
        filtered_generations.append(Chem.MolToSmiles(mol))
        filtered_scores.append(score)
    
    return filtered_generations, filtered_scores


def filter_single_atom(generations):
    filtered_generations = []
    for generation in generations:
        filtered_generations.append('.'.join([smi for smi in generation.split('.') if len(smi) > 1]))
    
    return filtered_generations


def translate(model, batch, args, vocab_tokens):
    with torch.no_grad():
        results = model.predict_step(
            reaction_batch=batch,
            batch_size=batch.size,
            beam_size=args.beam_size,
            n_best=args.expansion_topk,
            temperature=1,
            min_length=1,
            max_length=512
        )

        for predictions in results["predictions"]:
            smis = []
            for prediction in predictions:
                predicted_idx = prediction.detach().cpu().numpy()
                predicted_tokens = [vocab_tokens[idx] for idx in predicted_idx[:-1]]
                smi = "".join(predicted_tokens)
                smis.append(smi)
    
    generations, scores = smis, [s.cpu() for s in results['scores'][0]]

    generations, scores = filter_invalid(generations, scores)
    generations = filter_single_atom(generations)
    generations, scores = generations[:args.expansion_topk], scores[:args.expansion_topk]

    if len(scores) > 0:
        scores = list(np.exp(np.array(scores)) / np.sum(np.exp(np.array(scores))))   # softmax

    return {'reactants': generations, 'scores': scores}


def prepare_g2s(cuda=True, beam_size=200, expansion_topk=10):
    device = torch.device("cuda") if cuda else torch.device("cpu")
    # state = torch.load('g2s/saved_models/clean.pt')
    state = torch.load('g2s/saved_models/biochem.pt')
    args = state["args"]
    args.beam_size = beam_size
    args.expansion_topk = expansion_topk
    args.vocab_file = 'g2s/saved_models/vocab_smiles.txt'

    vocab = load_vocab(args.vocab_file)
    vocab_tokens = [k for k, v in sorted(vocab.items(), key=lambda tup: tup[1])]

    model = Graph2SeqSeriesRel(args, vocab)
    model.load_state_dict(state["state_dict"])
    model.to(device)
    model.eval()

    return model, args, vocab, vocab_tokens, device


def run_g2s(model, args, smi, vocab, vocab_tokens, device):
    batch = build_batch(smi, vocab)
    batch.to(device)
    result = translate(model, batch, args, vocab_tokens)
    result['templates'] = [None for _ in range(len(result['scores']))]
    result['retrieved'] = [False for _ in range(len(result['scores']))]
    return result


if __name__ == "__main__":
    t_start = time()
    model, args, vocab, vocab_tokens, device = prepare_g2s()
    # smi = 'N[C@@H](CNC(=O)C(=O)O)C(=O)O'
    smi = 'C=CC(C)(C)[C@@]12C[C@H]3c4nc5ccccc5c(=O)n4[C@H](C)C(=O)N3C1N(C(C)=O)c1ccccc12'
    print(run_g2s(model, args, smi, vocab, vocab_tokens, device))
    print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



import os
from time import time
import numpy as np
import argparse
import pandas as pd
import multiprocessing as mp
from utils.multithread_utils import *

mp.set_start_method('forkserver', force=True)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--products',     type=str, default='data/test.txt')
    parser.add_argument('-f', '--save_file',    type=str, default='ti/debug.txt')
    parser.add_argument('-b', '--blocks',       type=str, default='data/building_block.csv')
    parser.add_argument('-i', '--iterations',   type=int, default=500)
    parser.add_argument('-e', '--exp_topk',     type=int, default=10)
    parser.add_argument('-k', '--route_topk',   type=int, default=50)
    parser.add_argument('-s', '--beam_size',    type=int, default=10)
    parser.add_argument('-m', '--model_type', type=str, default='ensemble', choices=['ensemble','retroformer','g2s','retriever_only'])
    parser.add_argument('-r', '--retrieval',    type=str, default='true', choices=['true', 'false'])
    parser.add_argument('-pr', '--path_retrieval',    type=str, default='true', choices=['true', 'false'])
    parser.add_argument('-d', '--retrieval_db', type=str, default='data/train_canonicalized.txt')
    parser.add_argument('-pd', '--path_retrieval_db', type=str, default='data/pathways.pickle')
    parser.add_argument('-c', '--device',       type=str, default='cuda', choices=['cuda', 'cpu'])
    parser.add_argument('-n', '--num_threads',  type=int, default=10)
    args = parser.parse_args()

    t_start = time()

    if not os.path.exists(os.path.dirname(args.save_file)):
        os.makedirs(os.path.dirname(args.save_file))

    df = pd.read_csv(args.products, header=None, sep='\t')
    mol_ids, products = df[1], df[2]
    
    idx_split = np.array_split(range(len(products)), args.num_threads)
    idx_split = [(i[0], i[-1] + 1) for i in idx_split]
    print(idx_split)

    manager = mp.Manager()
    q = manager.Queue()
    file_pool = mp.Pool(1)
    file_pool.apply_async(listener, (q, args.save_file))

    pool = mp.Pool(args.num_threads)
    jobs = []
    for i in range(args.num_threads):
        start, end = idx_split[i]
        job = pool.apply_async(worker, (q, i, args, mol_ids[start:end], products[start:end]))
        jobs.append(job)

    for job in jobs:
        job.get()

    q.put('#done#')
    pool.close()
    pool.join()

    print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



import torch


def index_scatter(sub_data, all_data, index):
    d0, d1 = all_data.size()
    buf = torch.zeros_like(all_data).scatter_(0, index.repeat(d1, 1).t(), sub_data)
    mask = torch.ones(d0, device=all_data.device).scatter_(0, index, 0)

    return all_data * mask.unsqueeze(-1) + buf


def index_select_ND(source, dim, index):
    index_size = index.size()
    suffix_dim = source.size()[1:]
    final_size = index_size + suffix_dim
    target = source.index_select(dim, index.view(-1))

    return target.view(final_size)



import os
import pickle
import numpy as np
from rdkit import Chem

from torch.utils.data import Dataset
from retroformer.utils.smiles_utils import smi_tokenizer, clear_map_number, SmilesGraph


def gaussian(x, mean, amplitude, standard_deviation):
    return amplitude * np.exp(- (x - mean) ** 2 / (2 * standard_deviation ** 2))


class RetroDataset(Dataset):
    def __init__(self, smi, mode, data_folder='./data', vocab_file='./vocab_share.pk', augment=False):
        self.data_folder = data_folder

        assert mode in ['train', 'test', 'val']
        self.BondTypes = ['NONE', 'AROMATIC', 'DOUBLE', 'SINGLE', 'TRIPLE']
        self.bondtoi = {bond: i for i, bond in enumerate(self.BondTypes)}

        self.mode = mode
        self.augment = augment

        with open(vocab_file, 'rb') as f:
            self.src_itos, self.tgt_itos = pickle.load(f)
        self.src_stoi = {self.src_itos[i]: i for i in range(len(self.src_itos))}
        self.tgt_stoi = {self.tgt_itos[i]: i for i in range(len(self.tgt_itos))}

        self.processed = self.build_processed_data(smi)

        self.factor_func = lambda x: (1 + gaussian(x, 5.55391565, 0.27170542, 1.20071279)) # pre-computed

    def build_processed_data(self, smi):
        result = self.parse_smi(smi)
        if result is not None:
            src, src_graph = result
            return {'src': src, 'src_graph': src_graph, 'raw_product': smi}
        else:
            print('Warning. Process Failed.')
            return None

    def parse_smi(self, prod):
        cano_prod = clear_map_number(prod)
        if Chem.MolFromSmiles(cano_prod) is None:
            return None

        # Get the smiles graph
        smiles_graph = SmilesGraph(cano_prod)

        # Prepare model inputs
        src_token = smi_tokenizer(cano_prod)
        src_token = ['<UNK>'] + src_token
        src_token = [self.src_stoi.get(st, self.src_stoi['<unk>']) for st in src_token]

        return src_token, smiles_graph

    def reconstruct_smi(self, tokens, src=True, raw=False):
        if src:
            if raw:
                return [self.src_itos[t] for t in tokens]
            else:
                return [self.src_itos[t] for t in tokens if t != 1]
        else:
            if raw:
                return [self.tgt_itos[t] for t in tokens]
            else:
                return [self.tgt_itos[t] for t in tokens if t not in [1, 2, 3]]

    def __len__(self):
        return 1

    def __getitem__(self, idx):
        src, src_graph = self.processed['src'], self.processed['src_graph']
        src[0] = self.src_stoi['<UNK>']

        return src, src_graph



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from g2s.utils.chem_utils import ATOM_FDIM, BOND_FDIM
from g2s.utils.data_utils import G2SBatch
from g2s.models.attention_xl import AttnEncoderXL
from g2s.models.graphfeat import GraphFeatEncoder
from onmt.decoders import TransformerDecoder
from onmt.modules.embeddings import Embeddings
from onmt.translate import BeamSearch, GNMTGlobalScorer, GreedySearch
from typing import Any, Dict


class Graph2SeqSeriesRel(nn.Module):
    def __init__(self, args, vocab: Dict[str, int]):
        super().__init__()
        self.args = args
        self.vocab = vocab
        self.vocab_size = len(self.vocab)

        while args.enable_amp and not self.vocab_size % 8 == 0:
            self.vocab_size += 1

        self.encoder = GraphFeatEncoder(
            args,
            n_atom_feat=sum(ATOM_FDIM),
            n_bond_feat=BOND_FDIM
        )

        if args.attn_enc_num_layers > 0:
            self.attention_encoder = AttnEncoderXL(args)
        else:
            self.attention_encoder = None

        self.decoder_embeddings = Embeddings(
            word_vec_size=args.embed_size,
            word_vocab_size=self.vocab_size,
            word_padding_idx=self.vocab["_PAD"],
            position_encoding=True,
            dropout=args.dropout
        )

        self.decoder = TransformerDecoder(
            num_layers=args.decoder_num_layers,
            d_model=args.decoder_hidden_size,
            heads=args.decoder_attn_heads,
            d_ff=args.decoder_filter_size,
            copy_attn=False,
            self_attn_type="scaled-dot",
            dropout=args.dropout,
            attention_dropout=args.attn_dropout,
            embeddings=self.decoder_embeddings,
            max_relative_positions=args.max_relative_positions,
            aan_useffn=False,
            full_context_alignment=False,
            alignment_layer=-3,
            alignment_heads=0
        )

        if not args.attn_enc_hidden_size == args.decoder_hidden_size:
            self.bridge_layer = nn.Linear(args.attn_enc_hidden_size, args.decoder_hidden_size, bias=True)

        self.output_layer = nn.Linear(args.decoder_hidden_size, self.vocab_size, bias=True)

        self.criterion = nn.CrossEntropyLoss(
            ignore_index=self.vocab["_PAD"],
            reduction="mean"
        )

    def encode_and_reshape(self, reaction_batch: G2SBatch):
        hatom, _ = self.encoder(reaction_batch)                         # (n_atoms, h)
        if not self.args.attn_enc_hidden_size == self.args.decoder_hidden_size:
            hatom = self.bridge_layer(hatom)                            # bridging

        # hatom reshaping into [t, b, h]
        atom_scope = reaction_batch.atom_scope                          # list of b (n_components, 2)

        memory_lengths = [scope[-1][0] + scope[-1][1] - scope[0][0]
                          for scope in atom_scope]                      # (b, )

        # the 1+ corresponds to Atom(*)
        assert 1 + sum(memory_lengths) == hatom.size(0), \
            f"Memory lengths calculation error, encoder output: {hatom.size(0)}, memory_lengths: {memory_lengths}"

        memory_bank = torch.split(hatom, [1] + memory_lengths, dim=0)   # [n_atoms, h] => 1+b tup of (t, h)
        padded_memory_bank = []
        max_length = max(memory_lengths)

        for length, h in zip(memory_lengths, memory_bank[1:]):
            m = nn.ZeroPad2d((0, 0, 0, max_length - length))
            padded_memory_bank.append(m(h))

        padded_memory_bank = torch.stack(padded_memory_bank, dim=1)     # list of b (max_t, h) => [max_t, b, h]

        memory_lengths = torch.tensor(memory_lengths,
                                      dtype=torch.long,
                                      device=padded_memory_bank.device)

        if self.attention_encoder is not None:
            padded_memory_bank = self.attention_encoder(
                padded_memory_bank,
                memory_lengths,
                reaction_batch.distances
            )

        self.decoder.state["src"] = np.zeros(max_length)    # TODO: this is hardcoded to make transformer decoder work

        return padded_memory_bank, memory_lengths

    def forward(self, reaction_batch: G2SBatch):
        padded_memory_bank, memory_lengths = self.encode_and_reshape(reaction_batch)

        # adapted from onmt.models
        dec_in = reaction_batch.tgt_token_ids[:, :-1]                       # pop last, insert SOS for decoder input
        m = nn.ConstantPad1d((1, 0), self.vocab["_SOS"])
        dec_in = m(dec_in)
        dec_in = dec_in.transpose(0, 1).unsqueeze(-1)                       # [b, max_tgt_t] => [max_tgt_t, b, 1]

        dec_outs, _ = self.decoder(
            tgt=dec_in,
            memory_bank=padded_memory_bank,
            memory_lengths=memory_lengths
        )

        dec_outs = self.output_layer(dec_outs)                                  # [t, b, h] => [t, b, v]
        dec_outs = dec_outs.permute(1, 2, 0)                                    # [t, b, v] => [b, v, t]

        loss = self.criterion(
            input=dec_outs,
            target=reaction_batch.tgt_token_ids
        )

        predictions = torch.argmax(dec_outs, dim=1)                             # [b, t]
        mask = (reaction_batch.tgt_token_ids != self.vocab["_PAD"]).long()
        accs = (predictions == reaction_batch.tgt_token_ids).float()
        accs = accs * mask
        acc = accs.sum() / mask.sum()

        return loss, acc

    def predict_step(self, reaction_batch: G2SBatch,
                     batch_size: int, beam_size: int, n_best: int, temperature: float,
                     min_length: int, max_length: int) -> Dict[str, Any]:
        if beam_size == 1:
            decode_strategy = GreedySearch(
                pad=self.vocab["_PAD"],
                bos=self.vocab["_SOS"],
                eos=self.vocab["_EOS"],
                batch_size=batch_size,
                min_length=min_length,
                max_length=max_length,
                block_ngram_repeat=0,
                exclusion_tokens=set(),
                return_attention=False,
                sampling_temp=0.0,
                keep_topk=1
            )
        else:
            global_scorer = GNMTGlobalScorer(alpha=0.0,
                                             beta=0.0,
                                             length_penalty="none",
                                             coverage_penalty="none")
            decode_strategy = BeamSearch(
                beam_size=beam_size,
                batch_size=batch_size,
                pad=self.vocab["_PAD"],
                bos=self.vocab["_SOS"],
                eos=self.vocab["_EOS"],
                unk=self.vocab["_UNK"],
                ban_unk_token=True,
                n_best=n_best,
                global_scorer=global_scorer,
                min_length=min_length,
                max_length=max_length,
                return_attention=False,
                block_ngram_repeat=0,
                exclusion_tokens=set(),
                stepwise_penalty=None,
                ratio=0.0
            )

        padded_memory_bank, memory_lengths = self.encode_and_reshape(reaction_batch=reaction_batch)
        # adapted from onmt.translate.translator
        results = {
            "predictions": None,
            "scores": None,
            "attention": None
        }

        # (2) prep decode_strategy. Possibly repeat src objects.
        src_map = None
        target_prefix = None
        fn_map_state, memory_bank, memory_lengths, src_map = decode_strategy.initialize(
            memory_bank=padded_memory_bank,
            src_lengths=memory_lengths,
            src_map=src_map,
            target_prefix=target_prefix
        )

        # (3) Begin decoding step by step:
        for step in range(decode_strategy.max_length):
            decoder_input = decode_strategy.current_predictions.view(1, -1, 1)

            dec_out, dec_attn = self.decoder(tgt=decoder_input,
                                             memory_bank=memory_bank,
                                             memory_lengths=memory_lengths,
                                             step=step)

            if "std" in dec_attn:
                attn = dec_attn["std"]
            else:
                attn = None

            dec_out = self.output_layer(dec_out)            # [t, b, h] => [t, b, v]
            dec_out = dec_out / temperature
            dec_out = dec_out.squeeze(0)                    # [t, b, v] => [b, v]
            log_probs = F.log_softmax(dec_out, dim=-1)

            # log_probs = self.model.generator(dec_out.squeeze(0))

            decode_strategy.advance(log_probs, attn)
            any_finished = decode_strategy.is_finished.any()
            if any_finished:
                decode_strategy.update_finished()
                if decode_strategy.done:
                    break

            select_indices = decode_strategy.select_indices

            if any_finished:
                # Reorder states.
                if isinstance(memory_bank, tuple):
                    memory_bank = tuple(x.index_select(1, select_indices)
                                        for x in memory_bank)
                else:
                    memory_bank = memory_bank.index_select(1, select_indices)

                memory_lengths = memory_lengths.index_select(0, select_indices)

                if src_map is not None:
                    src_map = src_map.index_select(1, select_indices)

            if any_finished:
                self.map_state(
                    lambda state, dim: state.index_select(dim, select_indices))

        results["scores"] = decode_strategy.scores
        results["predictions"] = decode_strategy.predictions
        results["attention"] = decode_strategy.attention
        results["alignment"] = [[] for _ in range(self.args.predict_batch_size)]

        return results

    # adapted from onmt.decoders.transformer
    def map_state(self, fn):
        def _recursive_map(struct, batch_dim=0):
            for k, v in struct.items():
                if v is not None:
                    if isinstance(v, dict):
                        _recursive_map(v)
                    else:
                        struct[k] = fn(v, batch_dim)

        # self.decoder.state["src"] = fn(self.decoder.state["src"], 1)
        # => self.state["src"] = self.state["src"].index_select(1, select_indices)

        if self.decoder.state["cache"] is not None:
            _recursive_map(self.decoder.state["cache"])



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


import math
import torch
import torch.nn as nn
import numpy as np

from retroformer.models.encoder import TransformerEncoder
from retroformer.models.decoder import TransformerDecoder
from retroformer.models.embedding import Embedding
from retroformer.models.module import MultiHeadedAttention


class RetroModel(nn.Module):
    def __init__(self, encoder_num_layers, decoder_num_layers, d_model, heads, d_ff, dropout,
                 vocab_size_src, vocab_size_tgt, shared_vocab, num_bonds=5, shared_encoder=False, src_pad_idx=1,
                 tgt_pad_idx=1):
        super(RetroModel, self).__init__()
        self.encoder_num_layers = encoder_num_layers
        self.decoder_num_layers = decoder_num_layers
        self.d_model = d_model
        self.heads = heads
        self.d_ff = d_ff
        self.dropout = dropout
        self.shared_vocab = shared_vocab
        self.shared_encoder = shared_encoder
        if shared_vocab:
            assert vocab_size_src == vocab_size_tgt and src_pad_idx == tgt_pad_idx
            self.embedding_src = self.embedding_tgt = Embedding(vocab_size=vocab_size_src + 1, embed_size=d_model,
                                                                padding_idx=src_pad_idx)
        else:
            self.embedding_src = Embedding(vocab_size=vocab_size_src + 1, embed_size=d_model, padding_idx=src_pad_idx)
            self.embedding_tgt = Embedding(vocab_size=vocab_size_tgt + 1, embed_size=d_model, padding_idx=tgt_pad_idx)

        self.embedding_bond = nn.Linear(7, d_model)

        multihead_attn_modules_en = nn.ModuleList(
            [MultiHeadedAttention(heads, d_model, dropout=dropout)
             for _ in range(encoder_num_layers)])
        if shared_encoder:
            assert encoder_num_layers == decoder_num_layers
            multihead_attn_modules_de = multihead_attn_modules_en
        else:
            multihead_attn_modules_de = nn.ModuleList(
                [MultiHeadedAttention(heads, d_model, dropout=dropout)
                 for _ in range(decoder_num_layers)])

        self.encoder = TransformerEncoder(num_layers=encoder_num_layers,
                                          d_model=d_model, heads=heads,
                                          d_ff=d_ff, dropout=dropout,
                                          embeddings=self.embedding_src,
                                          embeddings_bond=self.embedding_bond,
                                          attn_modules=multihead_attn_modules_en)

        self.decoder = TransformerDecoder(num_layers=decoder_num_layers,
                                          d_model=d_model, heads=heads,
                                          d_ff=d_ff, dropout=dropout,
                                          embeddings=self.embedding_tgt,
                                          self_attn_modules=multihead_attn_modules_de)

        self.atom_rc_identifier = nn.Sequential(nn.Linear(d_model, 1),
                                                nn.Sigmoid())
        self.bond_rc_identifier = nn.Sequential(nn.Linear(d_model, 1),
                                                nn.Sigmoid())

        self.generator = nn.Sequential(nn.Linear(d_model, vocab_size_tgt),
                                       nn.LogSoftmax(dim=-1))

        self.softmax = nn.Softmax(dim=-1)

    def forward(self, src, tgt, bond=None, teacher_mask=None):
        encoder_out, edge_feature = self.encoder(src, bond)

        atom_rc_scores = self.atom_rc_identifier(encoder_out)
        bond_rc_scores = self.bond_rc_identifier(edge_feature) if edge_feature is not None else None

        if teacher_mask is None:  # Naive Inference
            student_mask = self.infer_reaction_center_mask(bond, atom_rc_scores, bond_rc_scores)
            decoder_out, top_aligns = self.decoder(src, tgt[:-1], encoder_out, student_mask.clone())
        else:  # Training
            decoder_out, top_aligns = self.decoder(src, tgt[:-1], encoder_out, teacher_mask.clone())

        generative_scores = self.generator(decoder_out)

        return generative_scores, atom_rc_scores, bond_rc_scores, top_aligns

    @staticmethod
    def infer_reaction_center_mask(bond, atom_rc_scores, bond_rc_scores=None):
        atom_rc_scores = atom_rc_scores.squeeze(2)
        if bond_rc_scores is not None:
            bond_rc_scores = bond_rc_scores.squeeze(1)
            bond_indicator = torch.zeros((bond.shape[0], bond.shape[1], bond.shape[2])).bool().to(bond.device)
            bond_indicator[bond.sum(-1) > 0] = (bond_rc_scores > 0.5)

            result = (~(bond_indicator.sum(dim=1).bool()) + ~(bond_indicator.sum(dim=2).bool()) +
                      (atom_rc_scores.transpose(0, 1) < 0.5)).transpose(0, 1)
        else:
            result = (atom_rc_scores.transpose(0, 1) < 0.5).transpose(0, 1)
        return result



import math
import argparse
from tqdm import tqdm
from rdkit import Chem
from retroformer.translate import prepare_retroformer, run_retroformer
from g2s.translate import prepare_g2s, run_g2s
from utils.ensemble import prepare_ensemble, run_ensemble_singlestep


def remove_chiral(smi):
    mol = Chem.MolFromSmiles(smi)
    if mol is None:
        return smi
    return Chem.MolToSmiles(mol, isomericSmiles=False)


def cal_acc(preds, target, n_best):
    correct_cnt = {'top1': 0, 'top3': 0, 'top5': 0, 'top10': 0}
    for i, tgt_smi in enumerate(target):
        pred_list = preds[i][:n_best]
        if tgt_smi in pred_list[:1]:
            correct_cnt['top1'] += 1
        if tgt_smi in pred_list[:3]:
            correct_cnt['top3'] += 1
        if tgt_smi in pred_list[:5]:
            correct_cnt['top5'] += 1
        if tgt_smi in pred_list[:10]:
            correct_cnt['top10'] += 1
    acc_dict = {key: value / len(target) for key, value in correct_cnt.items()}
    return acc_dict


parser = argparse.ArgumentParser()
parser.add_argument('-m', '--model_type',   type=str, default='ensemble', choices=['ensemble', 'retroformer', 'g2s'])
parser.add_argument('-b', '--batch_size',   type=int, default=32)
parser.add_argument('-s', '--beam_size',    type=int, default=10)
parser.add_argument('-d', '--device',       type=str, default='cuda', choices=['cuda', 'cpu'])
args = parser.parse_args()

if args.model_type == 'ensemble':
    model_retroformer, model_g2s, args_retroformer, args_g2s, vocab, vocab_tokens, device = \
        prepare_ensemble(args.device == 'cuda', args.beam_size)
elif args.model_type == 'retroformer':
    model_retroformer, args_retroformer = prepare_retroformer(args.device == 'cuda', args.beam_size)
else:
    model_g2s, args_g2s, vocab, vocab_tokens, device = prepare_g2s(args.device == 'cuda', args.beam_size)

with open('data/test_canonicalized_single.txt', 'r') as f:
    reactions = f.readlines()
tgt = [r.strip().split('>>')[0] for r in reactions]
src = [r.strip().split('>>')[1] for r in reactions]
tgt = [remove_chiral(t) for t in tgt]

preds = []
if args.model_type == 'ensemble':
    for i in tqdm(range(math.ceil(len(src) / args.batch_size))):
        smi_list = src[i * args.batch_size : (i + 1) * args.batch_size]
        preds.extend(run_ensemble_singlestep(smi_list, model_retroformer, model_g2s,
                                             args_retroformer, args_g2s,
                                             vocab, vocab_tokens, device))
elif args.model_type == 'retroformer':
    for i in tqdm(range(math.ceil(len(src) / args.batch_size))):
        smi_list = src[i * args.batch_size : (i + 1) * args.batch_size]
        preds.extend(run_retroformer(model_retroformer, args_retroformer, smi_list)[0])
else:
    for smi in tqdm(src):
        preds.append(run_g2s(model_g2s, args_g2s, smi, vocab, vocab_tokens, device)['reactants'])

for pred in preds:
    for i in range(len(pred)):
        pred[i] = remove_chiral(pred[i])

result = cal_acc(preds, tgt, n_best=10)
print(result)



from retro_star.api import RSPlanner
from time import time
import argparse

def run_retro_planner(product, blocks='data/building_block.csv', iterations=100, exp_topk=10, route_topk=10,
                      beam_size=10, model_type='ensemble', retrieval='true', path_retrieval='true',
                      retrieval_db='data/train_canonicalized.txt', path_retrieval_db='data/pathways.pickle',
                      device='cuda'):

    t_start = time()

    planner = RSPlanner(
        cuda=device == 'cuda',
        iterations=iterations,
        expansion_topk=exp_topk,
        route_topk=route_topk,
        beam_size=beam_size,
        model_type=model_type,
        retrieval=retrieval == 'true',
        retrieval_db=retrieval_db,
        path_retrieval=path_retrieval == 'true',
        path_retrieval_db=path_retrieval_db,
        starting_molecules=blocks
    )

    result = planner.plan(product)

    if result is None:
        print('None')
    else:
        for i, route in enumerate(result):
            print(f'{i} {route}')

    print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('product', type=str)
    parser.add_argument('-b', '--blocks', type=str, default='data/building_block.csv')
    parser.add_argument('-i', '--iterations', type=int, default=100)
    parser.add_argument('-e', '--exp_topk', type=int, default=10)
    parser.add_argument('-k', '--route_topk', type=int, default=10)
    parser.add_argument('-s', '--beam_size', type=int, default=10)
    parser.add_argument('-m', '--model_type', type=str, default='ensemble',
                        choices=['ensemble', 'retroformer', 'g2s', 'retriever_only'])
    parser.add_argument('-r', '--retrieval', type=str, default='true', choices=['true', 'false'])
    parser.add_argument('-pr', '--path_retrieval', type=str, default='true', choices=['true', 'false'])
    parser.add_argument('-d', '--retrieval_db', type=str, default='data/train_canonicalized.txt')
    parser.add_argument('-pd', '--path_retrieval_db', type=str, default='data/pathways.pickle')
    parser.add_argument('-c', '--device', type=str, default='cuda', choices=['cuda', 'cpu'])
    args = parser.parse_args()

    run_retro_planner(args.product, args.blocks, args.iterations, args.exp_topk, args.route_topk,
                      args.beam_size, args.model_type, args.retrieval, args.path_retrieval,
                      args.retrieval_db, args.path_retrieval_db, args.device)


import os
import sys
root = os.path.abspath('.')
if root not in sys.path:
    sys.path.insert(0, root)

from time import time
import numpy as np
from retroformer.translate import prepare_retroformer, run_retroformer
from g2s.translate import prepare_g2s, run_g2s


def prepare_ensemble(cuda=True, beam_size=10, expansion_topk=10):
    model_retroformer, args_retroformer = prepare_retroformer(cuda, beam_size, expansion_topk)
    model_g2s, args_g2s, vocab, vocab_tokens, device = prepare_g2s(cuda, 200, expansion_topk)
    return model_retroformer, model_g2s, args_retroformer, args_g2s, vocab, vocab_tokens, device


def run_ensemble(x, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                 vocab, vocab_tokens, device, expansion_topk=10):
    if model_type == 'ensemble':
        result = run_retroformer(model_retroformer, args_retroformer, x)
        _result = run_g2s(model_g2s, args_g2s, x, vocab, vocab_tokens, device)
        result['reactants'].extend(_result['reactants'])
        result['scores'].extend(_result['scores'])
        
    elif model_type == 'retroformer':
        result = run_retroformer(model_retroformer, args_retroformer, x)
    elif model_type == 'g2s':
        result = run_g2s(model_g2s, args_g2s, x, vocab, vocab_tokens, device)
    else:
        print("Wrong model type")   # Not the case
        
    reactants, scores = [], []
    for reactant, score in zip(result['reactants'], result['scores']):
        if reactant not in reactants:
            reactants.append(reactant)
            scores.append(score)
    result['reactants'], result['scores'] = reactants, scores
    
    ordering = np.argsort(result['scores'])[::-1][:expansion_topk]
    result['reactants'] = list(np.array(result['reactants'])[ordering])
    result['scores'] = list(np.array(result['scores'])[ordering])
    
    if len(result['scores']) > 0:
        result['scores'] = list(np.exp(np.array(result['scores'])) / np.sum(np.exp(np.array(result['scores']))))
    result['retrieved'] = [False for _ in result['scores']]
    result['templates'] = [None for _ in result['scores']]
    return result
        

def run_ensemble_singlestep(smi_list, model_retroformer, model_g2s, args_retroformer, args_g2s,
                            vocab, vocab_tokens, device):
    '''This is a batchrized version (for retroformer) of run_ensemble for faster evaluation'''
    reactants, scores = run_retroformer(model_retroformer, args_retroformer, smi_list)
    for i, smi in enumerate(smi_list):
        result = run_g2s(model_g2s, args_g2s, smi, vocab, vocab_tokens, device)
        reactants[i] = result['reactants'] + reactants[i]
        scores[i] = result['scores'] + scores[i]
    
    reactants_final = []
    for i in range(len(smi_list)):
        _reactants_final, _scores_final = [], []
        for reactant, score in zip(reactants[i], scores[i]):
            if reactant not in _reactants_final:
                _reactants_final.append(reactant)
                _scores_final.append(score)
        
        ordering = np.argsort(_scores_final)[::-1][:10]
        _reactants_final = list(np.array(_reactants_final)[ordering])
        reactants_final.append(_reactants_final)
    return reactants_final


if __name__ == '__main__':
    t_start = time()
    smi = 'CCCCCCCCCCCCCCCC(=O)OC(CCCCCCCCCCCCC)CC(=O)NCC(=O)O'
    model_retroformer, model_g2s, args_retroformer, args_g2s, vocab, vocab_tokens, device = prepare_ensemble()
    print(run_ensemble(smi, model_retroformer, model_g2s, args_retroformer, args_g2s, vocab, vocab_tokens, device))
    print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



from retro_star.api import RSPlanner
from time import time
import argparse


parser = argparse.ArgumentParser()
parser.add_argument('product',              type=str)
parser.add_argument('-b', '--blocks',       type=str, default='data/building_block.csv')
parser.add_argument('-i', '--iterations',   type=int, default=100)
parser.add_argument('-e', '--exp_topk',     type=int, default=10)
parser.add_argument('-k', '--route_topk',   type=int, default=10)
parser.add_argument('-s', '--beam_size',    type=int, default=10)
parser.add_argument('-m', '--model_type', type=str, default='ensemble', choices=['ensemble','retroformer','g2s','retriever_only'])
parser.add_argument('-r', '--retrieval',    type=str, default='true', choices=['true', 'false'])
parser.add_argument('-pr', '--path_retrieval',    type=str, default='true', choices=['true', 'false'])
parser.add_argument('-d', '--retrieval_db', type=str, default='data/train_canonicalized.txt')
parser.add_argument('-pd', '--path_retrieval_db', type=str, default='data/pathways.pickle')
parser.add_argument('-c', '--device',       type=str, default='cuda', choices=['cuda', 'cpu'])
args = parser.parse_args()

t_start = time()

planner = RSPlanner(
    cuda=args.device=='cuda',
    iterations=args.iterations,
    expansion_topk=args.exp_topk,
    route_topk=args.route_topk,
    beam_size=args.beam_size,
    model_type=args.model_type,
    retrieval=args.retrieval=='true',
    retrieval_db=args.retrieval_db,
    path_retrieval=args.path_retrieval=='true',
    path_retrieval_db=args.path_retrieval_db,
    starting_molecules=args.blocks
)

result = planner.plan(args.product)

if result is None:
    print('None')
else:
    for i, route in enumerate(result):
        print(f'{i} {route}')

print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



import torch
import torch.nn as nn
import numpy as np

from retroformer.models.module import PositionwiseFeedForward, LayerNorm, MultiHeadedAttention


class TransformerDecoderLayer(nn.Module):
    def __init__(self, d_model, heads, d_ff, dropout,
                 self_attn, context_attn):
        super(TransformerDecoderLayer, self).__init__()

        self.self_attn = self_attn
        self.context_attn = context_attn
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.layer_norm_1 = LayerNorm(d_model)
        self.layer_norm_2 = LayerNorm(d_model)
        self.dropout = dropout
        self.drop = nn.Dropout(dropout)
        mask = self._get_attn_subsequent_mask(5000)
        # Register self.mask as a buffer in TransformerDecoderLayer, so
        # it gets TransformerDecoderLayer's cuda behavior automatically.
        self.register_buffer('mask', mask)

    def forward(self, inputs, memory_bank, src_pad_mask, tgt_pad_mask,
                nonreactive_mask_input=None, layer_input=None, layer_cache=None):
        # inputs (`FloatTensor`): `[batch_size x 1 x model_dim]`
        # memory_bank (`FloatTensor`): `[batch_size x src_len x model_dim]`
        # src_pad_mask (`LongTensor`): `[batch_size x 1 x src_len]`
        # infer_decision_input (`LongTensor`): `[batch_size x tgt_len]`
        # nonreactive_mask_input (`BoolTensor`): `[batch_size x src_len]`
        dec_mask = torch.gt(tgt_pad_mask +
                            self.mask[:, :tgt_pad_mask.size(1),
                            :tgt_pad_mask.size(1)], 0)
        input_norm = self.layer_norm_1(inputs)

        # Self-attention:
        all_input = input_norm
        if layer_input is not None:
            all_input = torch.cat((layer_input, input_norm), dim=1)
            dec_mask = None
        query, self_attn, _ = self.self_attn(all_input, all_input, input_norm,
                                             mask=dec_mask,
                                             type="self",
                                             layer_cache=layer_cache)
        query = self.drop(query) + inputs
        query_norm = self.layer_norm_2(query)

        # Context-attention:
        mid, context_attn, _ = self.context_attn(memory_bank, memory_bank, query_norm,
                                                 mask=src_pad_mask,
                                                 additional_mask=nonreactive_mask_input,
                                                 type="context",
                                                 layer_cache=layer_cache)
        output = self.feed_forward(self.drop(mid) + query)

        return output, context_attn, all_input

    def _get_attn_subsequent_mask(self, size):
        attn_shape = (1, size, size)
        subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
        subsequent_mask = torch.from_numpy(subsequent_mask)
        return subsequent_mask


class TransformerDecoder(nn.Module):
    def __init__(self, num_layers, d_model, heads, d_ff, dropout, embeddings, self_attn_modules):
        super(TransformerDecoder, self).__init__()

        # Basic attributes.
        self.num_layers = num_layers
        self.embeddings = embeddings

        context_attn_modules = nn.ModuleList(
            [MultiHeadedAttention(heads, d_model, dropout=dropout)
             for _ in range(num_layers)])

        # Build TransformerDecoder.
        self.transformer_layers = nn.ModuleList(
            [TransformerDecoderLayer(d_model, heads, d_ff, dropout, self_attn_modules[i], context_attn_modules[i])
             for i in range(num_layers)])

        self.layer_norm_0 = LayerNorm(d_model)
        self.layer_norm = LayerNorm(d_model)

    def forward(self, src, tgt, memory_bank, nonreactive_mask=None, state_cache=None, step=None):
        '''
        :param src:
        :param tgt:
        :param memory_bank:
        :param nonreactive_mask: mask corresponding to reaction center identification from encoder
        :param infer_label: only occur in training for teacher's forcing; during inference, infer_label is the infer_decision.
        :param state_cache:
        :param step:
        :return:
        '''
        if nonreactive_mask is not None:
            nonreactive_mask[0] = False     # allow attention to the initial src token

        src_words = src.transpose(0, 1)
        tgt_words = tgt.transpose(0, 1)
        src_batch, src_len = src_words.size()
        tgt_batch, tgt_len = tgt_words.size()

        # Initialize return variables.
        outputs = []

        # Run the forward pass of the TransformerDecoder.
        emb = self.embeddings(tgt, step=step)
        assert emb.dim() == 3  # len x batch x embedding_dim
        if step is not None:
            tgt_words = tgt[-1].unsqueeze(0).transpose(0, 1)
            tgt_batch, tgt_len = tgt_words.size()

        output = emb.transpose(0, 1).contiguous()
        src_memory_bank = memory_bank.transpose(0, 1).contiguous()

        padding_idx = self.embeddings.word_padding_idx
        # assume src padding idx and tgt padding idx are the same
        src_pad_mask = src_words.data.eq(padding_idx).unsqueeze(1) \
            .expand(src_batch, tgt_len, src_len)
        tgt_pad_mask = tgt_words.data.eq(padding_idx).unsqueeze(1) \
            .expand(tgt_batch, tgt_len, tgt_len)

        nonreactive_mask_input = nonreactive_mask.transpose(0, 1) if nonreactive_mask is not None else None
        top_context_attns = []
        for i in range(self.num_layers):
            layer_input = None
            layer_cache = {'self_keys': None,
                           'self_values': None,
                           'memory_keys': None,
                           'memory_values': None}
            if state_cache is not None:
                layer_cache = state_cache.get('layer_cache_{}'.format(i), layer_cache)
                layer_input = state_cache.get('layer_input_{}'.format(i), layer_input)

            output, top_context_attn, all_input \
                = self.transformer_layers[i](
                    output, src_memory_bank,
                    src_pad_mask, tgt_pad_mask,
                    layer_input=layer_input,
                    layer_cache=layer_cache,
                    nonreactive_mask_input=nonreactive_mask_input)

            top_context_attns.append(top_context_attn)
            if state_cache is not None:
                state_cache['layer_cache_{}'.format(i)] = layer_cache
                state_cache['layer_input_{}'.format(i)] = all_input


        output = self.layer_norm(output)
        # Process the result and update the attentions.
        outputs = output.transpose(0, 1).contiguous()

        return outputs, top_context_attns



import numpy as np
from queue import Queue
from copy import deepcopy
from retro_star.alg.mol_node import MolNode
from retro_star.alg.reaction_node import ReactionNode
from retro_star.alg.syn_route import SynRoute


class MolTree:
    def __init__(self, target_mol, known_mols, value_fn, zero_known_value=True):
        self.target_mol = target_mol
        self.known_mols = known_mols
        self.value_fn = value_fn
        self.zero_known_value = zero_known_value
        self.mol_nodes = []
        self.reaction_nodes = []

        self.root = self._add_mol_node(target_mol, None)
        self.succ = target_mol in known_mols
        self.search_status = 0

        # if self.succ:
        #     logging.info('Synthesis route found: target in starting molecules')

    def _add_mol_node(self, mol, parent, retrieved=False):
        is_known = mol in self.known_mols

        init_value = self.value_fn(mol, retrieved)

        mol_node = MolNode(
            mol=mol,
            init_value=init_value,
            parent=parent,
            is_known=is_known,
            zero_known_value=self.zero_known_value
        )
        self.mol_nodes.append(mol_node)
        mol_node.id = len(self.mol_nodes)

        return mol_node

    def _add_reaction_and_mol_nodes(self, cost, mols, parent, template, ancestors, retrieved):
        assert cost >= 0

        for mol in mols:
            if mol in ancestors:
                return

        reaction_node = ReactionNode(parent, cost, template)
        for mol in mols:
            self._add_mol_node(mol, reaction_node, retrieved)
        reaction_node.init_values()
        self.reaction_nodes.append(reaction_node)
        reaction_node.id = len(self.reaction_nodes)

        return reaction_node

    def expand(self, mol_node, reactant_lists, costs, templates, retrieved):
        assert not mol_node.is_known and not mol_node.children

        if costs is None:      # No expansion results
            assert mol_node.init_values(no_child=True) == np.inf
            if mol_node.parent:
                mol_node.parent.backup(np.inf, from_mol=mol_node.mol)
            return self.succ

        assert mol_node.open
        ancestors = mol_node.get_ancestors()
        for i in range(len(costs)):
            self._add_reaction_and_mol_nodes(costs[i], reactant_lists[i],
                                             mol_node, templates[i], ancestors, retrieved[i])

        if len(mol_node.children) == 0:      # No valid expansion results
            assert mol_node.init_values(no_child=True) == np.inf
            if mol_node.parent:
                mol_node.parent.backup(np.inf, from_mol=mol_node.mol)
            return self.succ

        v_delta = mol_node.init_values()
        if mol_node.parent:
            mol_node.parent.backup(v_delta, from_mol=mol_node.mol)

        if not self.succ and self.root.succ:
            self.succ = True

        return self.succ

    def get_best_route(self):
        if not self.succ:
            return None

        syn_route = SynRoute(
            target_mol=self.root.mol,
            succ_value=self.root.succ_value,
            search_status=self.search_status
        )

        mol_queue = Queue()
        mol_queue.put(self.root)
        while not mol_queue.empty():
            mol = mol_queue.get()
            if mol.is_known:
                syn_route.set_value(mol.mol, mol.succ_value)
                continue

            best_reaction = None
            for reaction in mol.children:
                if reaction.succ:
                    if best_reaction is None or \
                            reaction.succ_value < best_reaction.succ_value:
                        best_reaction = reaction
            assert best_reaction.succ_value == mol.succ_value

            reactants = []
            for reactant in best_reaction.children:
                mol_queue.put(reactant)
                reactants.append(reactant.mol)

            syn_route.add_reaction(
                mol=mol.mol,
                value=mol.succ_value,
                template=best_reaction.template,
                reactants=reactants,
                cost=best_reaction.cost
            )

        return syn_route

    # from BioNavi-NP
    def get_routes(self):
        if not self.succ:
            return None

        routes = []
        syn_route = SynRoute(
            target_mol=self.root.mol,
            succ_value=self.root.succ_value,
            search_status=self.search_status
        )
        routes.append(syn_route)

        mol_queue = Queue()
        mol_queue.put((syn_route, self.root))
        while not mol_queue.empty():
            syn_route, mol = mol_queue.get()
            if mol.is_known:
                syn_route.set_value(mol.mol, mol.succ_value)
                continue

            best_reaction = None
            all_reactions = []
            for reaction in mol.children:
                if reaction.succ:
                    all_reactions.append(reaction)
                    if best_reaction is None or \
                            reaction.succ_value < best_reaction.succ_value:
                        best_reaction = reaction
            # assert best_reaction.succ_value == mol.succ_value

            syn_route_template = None
            if len(all_reactions) > 1:
                syn_route_template = deepcopy(syn_route)

            # best reaction
            reactants = []
            for reactant in best_reaction.children:
                mol_queue.put((syn_route, reactant))
                reactants.append(reactant.mol)

            syn_route.add_reaction(
                mol=mol.mol,
                value=mol.succ_value,
                template=best_reaction.template,
                reactants=reactants,
                cost=best_reaction.cost
            )

            # other reactions
            if len(all_reactions) > 1:
                for reaction in all_reactions:
                    if reaction == best_reaction:
                        continue

                    syn_route = deepcopy(syn_route_template)
                    routes.append(syn_route)

                    reactants = []
                    for reactant in reaction.children:
                        mol_queue.put((syn_route, reactant))
                        reactants.append(reactant.mol)

                    syn_route.add_reaction(
                        mol=mol.mol,
                        value=mol.succ_value,   # might be incorrect
                        template=reaction.template,
                        reactants=reactants,
                        cost=reaction.cost
                    )

            # print('succ value: %.2f | total cost: %.2f | length: %d' %
            #       (syn_route.succ_value, syn_route.total_cost, syn_route.length))

        return routes



import pickle
import pandas as pd
from retro_star.alg import molstar


def prepare_starting_molecules(filename):
    if filename[-3:] == 'csv':
        starting_mols = set(list(pd.read_csv(filename)['mol']))
    else:
        assert filename[-3:] == 'pkl'
        with open(filename, 'rb') as f:
            starting_mols = pickle.load(f)

    # logging.info('%d starting molecules loaded' % len(starting_mols))
    return starting_mols


def prepare_molstar_planner(expansion_handler, value_fn, starting_mols, iterations):
    plan_handle = lambda x: molstar(
        target_mol=x,
        starting_mols=starting_mols,
        expand_fn=expansion_handler,
        value_fn=value_fn,
        iterations=iterations,
    )
    
    return plan_handle



import torch
import torch.nn as nn
import torch.nn.functional as F

from retroformer.models.module import PositionwiseFeedForward, LayerNorm, MultiHeadedAttention


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, heads, d_ff, dropout, attn):
        super(TransformerEncoderLayer, self).__init__()

        self.self_attn = attn
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.layer_norm = LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, inputs, mask, edge_feature, pair_indices):
        input_norm = self.layer_norm(inputs)
        context, attn, edge_feature_updated = self.self_attn(input_norm, input_norm, input_norm, mask=mask,
                                                             edge_feature=edge_feature,
                                                             pair_indices=pair_indices)

        out = self.dropout(context) + inputs
        if edge_feature is not None:
            edge_feature = self.layer_norm(edge_feature + edge_feature_updated)
        return self.feed_forward(out), attn, edge_feature

class TransformerEncoder(nn.Module):
    def __init__(self, num_layers, d_model, heads, d_ff,
                 dropout, embeddings, embeddings_bond, attn_modules):
        super(TransformerEncoder, self).__init__()

        self.num_layers = num_layers
        self.embeddings = embeddings
        self.embeddings_bond = embeddings_bond
        self.transformer = nn.ModuleList(
            [TransformerEncoderLayer(d_model, heads, d_ff, dropout, attn_modules[i])
             for i in range(num_layers)])
        self.layer_norm = LayerNorm(d_model)


    def forward(self, src, bond=None):
        '''
        :param src: [src_len, batch_size]
        :param bond: [batch_size, src_len, src_len, 7]
        :return:
        '''
        global node_feature
        emb = self.embeddings(src)
        out = emb.transpose(0, 1).contiguous()

        if bond is not None:
            pair_indices = torch.where(bond.sum(-1) > 0)
            valid_bond = bond[bond.sum(-1) > 0]
            edge_feature = self.embeddings_bond(valid_bond.float())
        else:
            pair_indices, edge_feature = None, None

        words = src.transpose(0, 1)
        w_batch, w_len = words.size()
        padding_idx = self.embeddings.word_padding_idx
        mask = words.data.eq(padding_idx).unsqueeze(1) \
            .expand(w_batch, w_len, w_len)

        # Run the forward pass of every layer of the transformer.
        for i in range(self.num_layers):
            out, attn, edge_feature = self.transformer[i](out, mask, edge_feature, pair_indices)

        out = self.layer_norm(out)
        out = out.transpose(0, 1).contiguous()
        edge_out = self.layer_norm(edge_feature) if edge_feature is not None else None
        return out, edge_out




import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem

def smiles_to_fp(s, fp_dim=2048, pack=False):
    mol = Chem.MolFromSmiles(s)
    fp = AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=fp_dim)
    onbits = list(fp.GetOnBits())
    # import pdb; pdb.set_trace()
    arr = np.zeros(fp.GetNumBits(), dtype=bool)
    # arr = np.zeros(fp.GetNumBits(), dtype=np.bool)
    arr[onbits] = 1

    if pack:
        arr = np.packbits(arr)

    return arr

def batch_smiles_to_fp(s_list, fp_dim):
    fps = []
    for s in s_list:
        fps.append(smiles_to_fp(s, fp_dim))
    fps = np.array(fps)

    assert fps.shape[0] == len(s_list) and fps.shape[1] == fp_dim

    return fps


from tqdm import tqdm
from rdkit import Chem
import pandas as pd
import os, sys
import multiprocessing as mp
import numpy as np

def worker(q, thread_id, rootpath, target):
    msg = run_rp2(thread_id, rootpath, target)
    q.put(msg)

def listener(q):
    while True:
        m = q.get()
        if m == '#done#':
            break

def run_rp2(thread_id, rootpath, targets):
    print(f'[Thread {thread_id}] start ({len(targets)} targets)')
    if not os.path.exists(f"{rootpath}/paths"):
            os.mkdir(f'{rootpath}/paths')
    for target in targets:
        os.system(f"python -m rp2paths all {rootpath}/{target} --outdir {rootpath}/paths/{target.replace('.csv','')}")
    return(f'{target} done')

if __name__ == '__main__':
    root = '/home/taein/Retrosynthesis/rp2paths/examples'
    scope = os.listdir(root)
    scope = [i for i in scope if '_scope.csv' in i]
    num_threads = 1

    idx_split = np.array_split(range(len(scope)), num_threads)
    idx_split = [(i[0], i[-1] + 1) for i in idx_split]

    manager = mp.Manager()
    q = manager.Queue()
    file_pool = mp.Pool(1)
    file_pool.apply_async(listener, (q))

    pool = mp.Pool(num_threads)
    jobs = []
    for i in range(num_threads):
        start, end = idx_split[i]
        job = pool.apply_async(worker, (q, i, root, scope[start:end]))
        jobs.append(job)

    for job in jobs:
        job.get()

    q.put('#done#')
    pool.close()
    pool.join()


import torch
import pandas as pd
from retro_star.common import prepare_starting_molecules, \
     prepare_molstar_planner, smiles_to_fp
from retro_star.model import ValueMLP
from retro_star.utils import setup_logger
from retroformer.translate import prepare_retroformer, run_retroformer
from utils.ensemble import prepare_ensemble, run_ensemble, prepare_g2s, run_g2s
from retro_star.retriever import Retriever, run_retriever, run_retriever_only, \
    neutralize_atoms, kegg_search, \
    pathRetriever, run_path_retriever, run_both_retriever
from rdkit import Chem


class RSPlanner:
    def __init__(self,
                 cuda=True,
                 beam_size=10,
                 iterations=500,
                 expansion_topk=10,
                 use_value_fn=True,
                 starting_molecules='data/building_block.csv',
                 value_model='retro_star/saved_model/best_epoch_final_4.pt',
                 model_type='ensemble',
                 retrieval_db='data/train_canonicalized.txt',
                 path_retrieval_db='/home/taein/df',
                 kegg_mol_db = "data/kegg_neutral_iso_smi.csv",
                 route_topk=10,
                 retrieval=True,
                 path_retrieval=True):
        
        self.path_retrieve_token = 'keggpath'
        self.kegg_mol_db = pd.read_csv(kegg_mol_db)
        self.path_retrieval_db = pd.read_pickle(path_retrieval_db)
        
        setup_logger()
        device = torch.device('cuda' if cuda else 'cpu')
        # device = torch.device('cuda:6' if cuda else 'cpu')
        starting_mols = prepare_starting_molecules(starting_molecules)

        # model_retroformer, args = prepare_retroformer(cuda, beam_size, expansion_topk)
        # model_g2s, args_g2s, vocab, vocab_tokens, _ = prepare_g2s(cuda, beam_size, expansion_topk)
        model_retroformer, model_g2s, args_retroformer, args_g2s, vocab, vocab_tokens, _ = \
            prepare_ensemble(cuda, beam_size, expansion_topk)
            
        if model_type == "retriever_only":
            starting_mols.add(self.path_retrieve_token)
            path_retriever = pathRetriever(kegg_mol_db, path_retrieval_db, self.path_retrieve_token)
            retriever = Retriever(retrieval_db)
            expansion_handler = lambda x: run_retriever_only(x, path_retriever, retriever)
                    
        elif path_retrieval and retrieval:
            starting_mols.add(self.path_retrieve_token)
            path_retriever = pathRetriever(kegg_mol_db, path_retrieval_db, self.path_retrieve_token)
            retriever = Retriever(retrieval_db)
            expansion_handler = lambda x: run_both_retriever(x, path_retriever, retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                                        vocab, vocab_tokens, device, expansion_topk)
        
        elif path_retrieval and not retrieval:
            starting_mols.add(self.path_retrieve_token)
            path_retriever = pathRetriever(kegg_mol_db, path_retrieval_db, self.path_retrieve_token)
            expansion_handler = lambda x: run_path_retriever(x, path_retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                                        vocab, vocab_tokens, device, expansion_topk)
        
        elif not path_retrieval and retrieval:
            retriever = Retriever(retrieval_db)
            expansion_handler = lambda x: run_retriever(x, retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                                        vocab, vocab_tokens, device, expansion_topk)
        else:
            expansion_handler = lambda x: run_ensemble(x, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                                       vocab, vocab_tokens, device, expansion_topk)

        
        self.top_k = route_topk

        if use_value_fn:
            model = ValueMLP(
                n_layers=1,
                fp_dim=2048,
                latent_dim=128,
                dropout_rate=0.1,
                device=device
            ).to(device)
            model.load_state_dict(torch.load(value_model, map_location=device))
            model.eval()

            def value_fn(mol, retrieved):
                if retrieved: return 0.
                # import pdb; pdb.set_trace()
                fp = smiles_to_fp(mol, fp_dim=2048).reshape(1, -1)
                fp = torch.FloatTensor(fp).to(device)
                v = model(fp).item()
                return v
        else:
            value_fn = lambda x, r: 0.

        self.plan_handle = prepare_molstar_planner(
            expansion_handler=expansion_handler,
            value_fn=value_fn,
            starting_mols=starting_mols,
            iterations=iterations
        )
    def __keggpath_find(self,routes,token,mol_db,path_db,top_k):
        modi_list = []
        for route in routes[:top_k]:
            r = route.split(">")
            token_position = [i for i,j in enumerate(r) if token in j]
            for pos in token_position:
                cid, _ = kegg_search(neutralize_atoms(r[pos-2].split("|")[-1]),mol_db)
                target_maps = path_db["Map"][path_db['Pathways'].apply(lambda x: any(cid in sublist for sublist in x))].to_list()
                map = target_maps[0]  # check a representation method
                r[pos] = r[pos].replace(token,f'{token}=kegg.jp/pathway/{map}+{cid}')
                if target_maps == []:  # not the case
                    modi_list.append(route)

            modi_route = '>'.join(r)
            modi_list.append(modi_route)
        return modi_list
            
    
    def plan(self, target_mol):
        try:
            target_mol = Chem.MolToSmiles(Chem.MolFromSmiles(target_mol))
            succ, msg = self.plan_handle(target_mol)    # the result of model
            
            if succ:
                routes_list = []
                for route in msg:
                    routes_list.append(route.serialize())
                modi_routes_list = self.__keggpath_find(routes_list,self.path_retrieve_token,self.kegg_mol_db,self.path_retrieval_db,self.top_k)
                return modi_routes_list
            
            elif target_mol != Chem.MolToSmiles(Chem.MolFromSmiles(target_mol),isomericSmiles=False):
                target_mol = Chem.MolToSmiles(Chem.MolFromSmiles(target_mol),isomericSmiles=False)
                succ, msg = self.plan_handle(target_mol)    # the result of model
                if succ:
                    routes_list = []
                    for route in msg:
                        routes_list.append(route.serialize())
                    modi_routes_list = self.__keggpath_find(routes_list,self.path_retrieve_token,self.kegg_mol_db,self.path_retrieval_db,self.top_k)
                    return modi_routes_list
                else:
                    return None
            else:
                return None
        except:
            return None


import math
import torch
import torch.nn as nn
from onmt.modules.embeddings import PositionalEncoding
from onmt.modules.position_ffn import PositionwiseFeedForward
from onmt.utils.misc import sequence_mask


def get_sin_encodings(rel_pos_buckets, model_dim) -> torch.Tensor:
    pe = torch.zeros(rel_pos_buckets + 1, model_dim)
    position = torch.arange(0, rel_pos_buckets).unsqueeze(1)
    div_term = torch.exp((torch.arange(0, model_dim, 2, dtype=torch.float) *
                          -(math.log(10000.0) / model_dim)))
    pe[:-1, 0::2] = torch.sin(position.float() * div_term)          # leaving last "position" as padding
    pe[:-1, 1::2] = torch.cos(position.float() * div_term)

    return pe


class MultiHeadedRelAttention(nn.Module):
    def __init__(self, args, head_count, model_dim, dropout, rel_pos_buckets, u, v):
        super().__init__()
        self.args = args

        assert model_dim % head_count == 0
        self.dim_per_head = model_dim // head_count
        self.model_dim = model_dim
        self.head_count = head_count

        self.linear_keys = nn.Linear(model_dim, model_dim)
        self.linear_values = nn.Linear(model_dim, model_dim)
        self.linear_query = nn.Linear(model_dim, model_dim)

        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.final_linear = nn.Linear(model_dim, model_dim)

        self.rel_pos_buckets = rel_pos_buckets

        if args.rel_pos == "enc_only":
            self.relative_pe = nn.Embedding.from_pretrained(
                embeddings=get_sin_encodings(rel_pos_buckets, model_dim),
                freeze=True,
                padding_idx=rel_pos_buckets
            )
            # self.W_kR = nn.Parameter(
            #     torch.Tensor(self.head_count, self.dim_per_head, self.dim_per_head), requires_grad=True)
            # self.b_kR = nn.Parameter(
            #     torch.Tensor(self.head_count, self.dim_per_head), requires_grad=True)

        elif args.rel_pos == "emb_only":
            self.relative_pe = nn.Embedding(
                rel_pos_buckets + 1,
                model_dim,
                padding_idx=rel_pos_buckets
            )
            # self.W_kR = nn.Parameter(
            #     torch.Tensor(self.head_count, self.dim_per_head, self.dim_per_head), requires_grad=True)
            # self.b_kR = nn.Parameter(
            #     torch.Tensor(self.head_count, self.dim_per_head), requires_grad=True)

        else:
            self.relative_pe = None
            self.W_kR = None
            self.b_kR = None

        self.u = u
        self.v = v

    def forward(self, inputs, mask, distances):
        """
        Compute the context vector and the attention vectors.

        Args:
           inputs (FloatTensor): set of `key_len`
               key vectors ``(batch, key_len, dim)``
           mask: binary mask 1/0 indicating which keys have
               zero / non-zero attention ``(batch, query_len, key_len)``
           distances: graph distance matrix (BUCKETED), ``(batch, key_len, key_len)``
        Returns:
           (FloatTensor, FloatTensor):

           * output context vectors ``(batch, query_len, dim)``
           * Attention vector in heads ``(batch, head, query_len, key_len)``.
        """

        batch_size = inputs.size(0)
        dim_per_head = self.dim_per_head
        head_count = self.head_count

        def shape(x):
            """Projection."""
            return x.view(batch_size, -1, head_count, dim_per_head).transpose(1, 2)

        def unshape(x):
            """Compute context."""
            return x.transpose(1, 2).contiguous().view(batch_size, -1, head_count * dim_per_head)

        # 1) Project key, value, and query. Seems that we don't need layer_cache here
        query = self.linear_query(inputs)
        key = self.linear_keys(inputs)
        value = self.linear_values(inputs)

        key = shape(key)                # (b, t_k, h) -> (b, head, t_k, h/head)
        value = shape(value)
        query = shape(query)            # (b, t_q, h) -> (b, head, t_q, h/head)

        key_len = key.size(2)
        query_len = query.size(2)

        # 2) Calculate and scale scores.
        query = query / math.sqrt(dim_per_head)

        if self.relative_pe is None:
            scores = torch.matmul(
                query, key.transpose(2, 3))                 # (b, head, t_q, t_k)

        else:
            # a + c
            u = self.u.reshape(1, head_count, 1, dim_per_head)
            a_c = torch.matmul(query + u, key.transpose(2, 3))

            rel_emb = self.relative_pe(distances)           # (b, t_q, t_k) -> (b, t_q, t_k, h)
            rel_emb = rel_emb.reshape(                      # (b, t_q, t_k, h) -> (b, t_q, t_k, head, h/head)
                batch_size, query_len, key_len, head_count, dim_per_head)

            # W_kR = self.W_kR.reshape(1, 1, 1, head_count, dim_per_head, dim_per_head)
            # rel_emb = torch.matmul(rel_emb, W_kR)           # (b, t_q, t_k, head, 1, h/head)
            # rel_emb = rel_emb.squeeze(-2)                   # (b, t_q, t_k, head, h/head)
            #
            # b_kR = self.b_kR.reshape(1, 1, 1, head_count, dim_per_head)
            # rel_emb = rel_emb + b_kR                        # (b, t_q, t_k, head, h/head)

            # b + d
            query = query.unsqueeze(-2)                     # (b, head, t_q, h/head) -> (b, head, t_q, 1, h/head)
            rel_emb_t = rel_emb.permute(0, 3, 1, 4, 2)      # (b, t_q, t_k, head, h/head) -> (b, head, t_q, h/head, t_k)

            v = self.v.reshape(1, head_count, 1, 1, dim_per_head)
            b_d = torch.matmul(query + v, rel_emb_t
                               ).squeeze(-2)                # (b, head, t_q, 1, t_k) -> (b, head, t_q, t_k)

            scores = a_c + b_d

        scores = scores.float()

        mask = mask.unsqueeze(1)                            # (B, 1, 1, T_values)
        scores = scores.masked_fill(mask, -1e18)

        # 3) Apply attention dropout and compute context vectors.
        attn = self.softmax(scores)
        drop_attn = self.dropout(attn)

        context_original = torch.matmul(drop_attn, value)   # -> (b, head, t_q, h/head)
        context = unshape(context_original)                 # -> (b, t_q, h)

        output = self.final_linear(context)
        attns = attn.view(batch_size, head_count, query_len, key_len)

        return output, attns


class SALayerXL(nn.Module):
    """
    A single layer of the self-attention encoder.

    Args:
        d_model (int): the dimension of keys/values/queries in
                   MultiHeadedAttention, also the input size of
                   the first-layer of the PositionwiseFeedForward.
        heads (int): the number of head for MultiHeadedAttention.
        d_ff (int): the second-layer of the PositionwiseFeedForward.
        dropout: dropout probability(0-1.0).
    """

    def __init__(self, args, d_model, heads, d_ff, dropout, attention_dropout, rel_pos_buckets: int, u, v):
        super().__init__()

        self.self_attn = MultiHeadedRelAttention(
            args,
            heads, d_model, dropout=attention_dropout,
            rel_pos_buckets=rel_pos_buckets,
            u=u,
            v=v
        )
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)
        self.dropout = nn.Dropout(dropout)

    def forward(self, inputs, mask, distances):
        """
        Args:
            inputs (FloatTensor): ``(batch_size, src_len, model_dim)``
            mask (LongTensor): ``(batch_size, 1, src_len)``
            distances (LongTensor): ``(batch_size, src_len, src_len)``

        Returns:
            (FloatTensor):

            * outputs ``(batch_size, src_len, model_dim)``
        """
        input_norm = self.layer_norm(inputs)
        context, _ = self.self_attn(input_norm, mask=mask, distances=distances)
        out = self.dropout(context) + inputs

        return self.feed_forward(out)


class AttnEncoderXL(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args

        self.num_layers = args.attn_enc_num_layers
        self.d_model = args.attn_enc_hidden_size
        self.heads = args.attn_enc_heads
        self.d_ff = args.attn_enc_filter_size
        self.attention_dropout = args.attn_dropout
        self.rel_pos_buckets = args.rel_pos_buckets

        self.encoder_pe = None
        if args.encoder_positional_encoding == "transformer":
            self.encoder_pe = PositionalEncoding(
                dropout=args.dropout,
                dim=self.d_model,
                max_len=1024        # temporary hard-code. Seems that onmt fix the denominator as 10000.0
            )
        else:
            self.dropout = nn.Dropout(p=args.dropout)

        if args.rel_pos in ["enc_only", "emb_only"]:
            self.u = nn.Parameter(torch.randn(self.d_model), requires_grad=True)
            self.v = nn.Parameter(torch.randn(self.d_model), requires_grad=True)
        else:
            self.u = None
            self.v = None

        self.attention_layers = nn.ModuleList(
            [SALayerXL(
                args, self.d_model, self.heads, self.d_ff, args.dropout, self.attention_dropout,
                self.rel_pos_buckets, self.u, self.v)
             for i in range(self.num_layers)])
        self.layer_norm = nn.LayerNorm(self.d_model, eps=1e-6)

    def forward(self, src, lengths, distances):
        """adapt from onmt TransformerEncoder
            src: (t, b, h)
            lengths: (b,)
            distances: (b, t, t)
        """

        if self.encoder_pe is not None:
            emb = self.encoder_pe(src)
            out = emb.transpose(0, 1).contiguous()
        else:
            out = src.transpose(0, 1).contiguous()
            if self.args.encoder_emb_scale == "sqrt":
                out = out * math.sqrt(self.d_model)
            out = self.dropout(out)

        mask = ~sequence_mask(lengths).unsqueeze(1)

        for layer in self.attention_layers:
            out = layer(out, mask, distances)
        out = self.layer_norm(out)

        return out.transpose(0, 1).contiguous()



import os
import numpy as np
import logging
from retro_star.alg.mol_tree import MolTree


def molstar(target_mol, starting_mols, expand_fn, value_fn, iterations):
    mol_tree = MolTree(
        target_mol=target_mol,
        known_mols=starting_mols,
        value_fn=value_fn
    )

    if not mol_tree.succ:
        for i in range(iterations):
            print(i)
            scores = []
            for m in mol_tree.mol_nodes:
                if m.open:
                    scores.append(m.v_target())
                else:
                    scores.append(np.inf)
            scores = np.array(scores)

            if np.min(scores) == np.inf:
                # logging.info('No open nodes!')
                break

            metric = scores

            mol_tree.search_status = np.min(metric)
            m_next = mol_tree.mol_nodes[np.argmin(metric)]
            assert m_next.open

            result = expand_fn(m_next.mol)
            print(result)
            if result is not None and len(result['scores']) > 0:
                reactants = result['reactants']
                scores = result['scores']
                retrieved = result['retrieved']
                costs = 0.0 - np.log(np.clip(np.array(scores), 1e-3, 1.0))
                # import pdb; pdb.set_trace()

                if 'templates' in result.keys():
                    templates = result['templates']
                else:
                    import pdb; pdb.set_trace()
                    templates = result['templates'] # need to fix for integrity

                reactant_lists = []
                for j in range(len(scores)):
                    reactant_list = list(set(reactants[j].split('.')))
                    reactant_lists.append(reactant_list)

                assert m_next.open
                
                # from BioNavi-NP
                mol_tree.expand(m_next, reactant_lists, costs, templates, retrieved)

            else:
                mol_tree.expand(m_next, None, None, None, None)

    routes = []
    if mol_tree.succ:
        # import pdb; pdb.set_trace()
        best_route = mol_tree.get_best_route()
        routes = mol_tree.get_routes()
        routes = sorted(routes, key=lambda x: x.total_cost)
        assert best_route is not None

    return mol_tree.succ, routes



import numpy as np
import logging


class ReactionNode:
    def __init__(self, parent, cost, template):
        self.parent = parent
        
        self.depth = self.parent.depth + 1
        self.id = -1

        self.cost = cost
        self.template = template
        self.children = []
        self.value = None   # [V(m | subtree_m) for m in children].sum() + cost
        self.succ_value = np.inf    # total cost for existing solution
        self.target_value = None    # V_target(self | whole tree)
        self.succ = None    # successfully found a valid synthesis route
        self.open = True    # before expansion: True, after expansion: False
        parent.children.append(self)

    def v_self(self):
        """
        :return: V_self(self | subtree)
        """
        return self.value

    def v_target(self):
        """
        :return: V_target(self | whole tree)
        """
        return self.target_value

    def init_values(self):
        assert self.open

        self.value = self.cost
        self.succ = True
        for mol in self.children:
            self.value += mol.value
            self.succ &= mol.succ

        if self.succ:
            self.succ_value = self.cost
            for mol in self.children:
                self.succ_value += mol.succ_value

        self.target_value = self.parent.v_target() - self.parent.v_self() + \
                            self.value
        self.open = False

    def backup(self, v_delta, from_mol=None):
        self.value += v_delta
        self.target_value += v_delta

        self.succ = True
        for mol in self.children:
            self.succ &= mol.succ

        if self.succ:
            self.succ_value = self.cost
            for mol in self.children:
                self.succ_value += mol.succ_value

        if v_delta != 0:
            assert from_mol
            self.propagate(v_delta, exclude=from_mol)

        return self.parent.backup(self.succ)

    def propagate(self, v_delta, exclude=None):
        if exclude is None:
            self.target_value += v_delta

        for child in self.children:
            if exclude is None or child.mol != exclude:
                for grandchild in child.children:
                    grandchild.propagate(v_delta)

    def serialize(self):
        return '%d' % (self.id)
        # return '%d | value %.2f | target %.2f' % \
        #        (self.id, self.v_self(), self.v_target())


from __future__ import print_function
import sys
import os
import re
import copy

import rdkit.Chem as Chem
import rdkit.Chem.AllChem as AllChem
from rdkit.Chem.rdchem import ChiralType, BondType, BondDir

from rdchiral.utils import vprint, PLEVEL, atoms_are_different
from rdchiral.initialization import rdchiralReaction, rdchiralReactants
from rdchiral.chiral import template_atom_could_have_been_tetra, copy_chirality,\
    atom_chirality_matches
from rdchiral.clean import canonicalize_outcome_smiles, combine_enantiomers_into_racemic
from rdchiral.bonds import BondDirOpposite, restore_bond_stereo_to_sp2_atom

'''
This file contains the main functions for running reactions.

An incomplete description of expected behavior is as follows:

(1) RDKit's native RunReactants is called on an achiral version of the molecule,
which has had all tetrahedral centers and bond directions stripped.

(2) For each outcome, we examine the correspondence between atoms in the
reactants and atoms in the reactant template for reasons to exclude the
current outcome. The way we do so is through the react_atom_idx property in
the generated products. This is one of the
few properties always copied over to the products here:
https://github.com/rdkit/rdkit/blob/master/Code/GraphMol/ChemReactions/ReactionRunner.cpp

A previous version of this code did so through the Isotope label of each atom,
before the react_atom_idx was added to the ReactionRunner.cpp code.

The following conditions are checked:

    TETRAHEDRAL ATOMS
    (a) If a reactant atom is a tetrahedral center with specified chirality
        and the reactant template atom is NOT chiral but is defined in a way
        that it could have been specified, reject this outcome
    (b) If a reactant atom is a tetrahedral center with specified chirality
        and the reactant template atom is NOT chiral and is not defined in
        a way where it could have been (i.e., is generalized without spec.
        neighbors), then keep the match.
    (c) If a reactant atom is achiral but the reactant tempalte atom is chiral,
        the match is still allowed to happen. We might want to change this later
        or let it be an option.
    (d) If a reactant atom is a tetrahedral center with specified chirality
        and the reactant template also has its chirality specified, let the
        match happen if the chirality matches.


    DOUBLE BONDS
    (a) If a reactant double bond is defined with directionality specified and
        the reactant template is unspecified but COULD have been (i.e.,
        neighbors of sp2 carbons are specified), reject this outcome
    (b) If a reactant double bond is defined with directionality specified and
        the reactant template si unspecified but could NOT have been (in the
        case of generalization), allow the match to occur. This is what we
        default to when half the bond is specified, like in "C=C/O"
    note: reactants are checked for implicit bond stereo based on rings
    (c) If a reactant double bond has implicit cis due to ring membership, it is
        still allowed to match an unspecified template double bond. Might lead
        to some weird edge cases, but mostly makes sense.


(3) For each outcome, merge all products into a single molecule. During this
process, we check for bonds that are missing in the product. These are those
that were present in the reactants but were NOT matched in the reactant
template.

(4) For each outcome, examine product atoms to correct tetrahedral chirality.

(5) For each outcome, examine product double bonds to correct cis/trans-ness

'''

def rdchiralRunText(reaction_smarts, reactant_smiles, **kwargs):
    '''Run from SMARTS string and SMILES string. This is NOT recommended
    for library application, since initialization is pretty slow. You should
    separately initialize the template and molecules and call run()'''
    rxn = rdchiralReaction(reaction_smarts)
    reactants = rdchiralReactants(reactant_smiles)
    return rdchiralRun(rxn, reactants, **kwargs)

def rdchiralRun(rxn, reactants, keep_mapnums=False, combine_enantiomers=True, return_mapped=False):
    '''
    rxn = rdchiralReaction (rdkit reaction + auxilliary information)
    reactants = rdchiralReactants (rdkit mol + auxilliary information)

    note: there is a fair amount of initialization (assigning stereochem), most
    importantly assigning atom map numbers to the reactant atoms. It is
    HIGHLY recommended to use the custom classes for initialization.
    '''

    # New: reset atom map numbers for templates in case they have been overwritten
    # by previous uses of this template!
    rxn.reset()

    ###############################################################################
    # Run naive RDKit on ACHIRAL version of molecules
    outcomes = rxn.rxn.RunReactants((reactants.reactants_achiral,))
    if PLEVEL >= (1): print('Using naive RunReactants, {} outcomes'.format(len(outcomes)))
    if not outcomes:
        return []
    ###############################################################################


    ###############################################################################
    # Initialize, now that there is at least one outcome

    final_outcomes = set()
    mapped_outcomes = {}
    # We need to keep track of what map numbers correspond to which atoms
    # note: all reactant atoms must be mapped, so this is safe
    atoms_r = reactants.atoms_r

    # Copy reaction template so we can play around with map numbers
    template_r, template_p = rxn.template_r, rxn.template_p

    # Get molAtomMapNum->atom dictionary for tempalte reactants and products
    atoms_rt_map = rxn.atoms_rt_map
    # TODO: cannot change atom map numbers in atoms_rt permanently?
    atoms_pt_map = rxn.atoms_pt_map
    ###############################################################################


    for outcome in outcomes:

        ###############################################################################
        # Look for new atoms in products that were not in
        # reactants (e.g., LGs for a retro reaction)
        if PLEVEL >= (2): print('Processing {}'.format(str([Chem.MolToSmiles(x, True) for x in outcome])))
        unmapped = 900
        for m in outcome:
            for a in m.GetAtoms():
                # Assign map number to outcome based on react_atom_idx
                if a.HasProp('react_atom_idx'):
                    a.SetAtomMapNum(reactants.idx_to_mapnum(int(a.GetProp('react_atom_idx'))))
                if not a.GetAtomMapNum():
                    a.SetAtomMapNum(unmapped)
                    unmapped += 1
        if PLEVEL >= 2: print('Added {} map numbers to product'.format(unmapped-900))
        ###############################################################################


        ###############################################################################
        # Check to see if reactants should not have been matched (based on chirality)

        # Define map num -> reactant template atom map
        atoms_rt =  {a.GetAtomMapNum(): atoms_rt_map[a.GetIntProp('old_mapno')] \
            for m in outcome for a in m.GetAtoms() if a.HasProp('old_mapno')}

        # Set map numbers of reactant template to be consistent with reactant/product molecules
        # note: this is okay to do within the loop, because ALL atoms must be matched
        # in the templates, so the atommapnum will get overwritten every time
        [a.SetAtomMapNum(i) for (i, a) in atoms_rt.items()]

        # Make sure each atom matches
        # note: this is a little weird because atom_chirality_matches takes three values,
        #       -1 (both tetra but opposite), 0 (not a match), and +1 (both tetra and match)
        #       and we only want to continue if they all equal -1 or all equal +1
        prev = None
        skip_outcome = False
        for match in (atom_chirality_matches(atoms_rt[i], atoms_r[i]) for i in atoms_rt):
            if match == 0:
                if PLEVEL >= 2: print('Chirality violated! Should not have gotten this match')
                skip_outcome = True
                break
            elif match == 2: # ambiguous case
                continue
            elif prev is None:
                prev = match
            elif match != prev:
                if PLEVEL >= 2: print('Part of the template matched reactant chirality, part is inverted! Should not match')
                skip_outcome = True
                break
        if skip_outcome:
            continue
        if PLEVEL >= 2: print('Chirality matches! Just checked with atom_chirality_matches')

        # Check bond chirality - iterate through reactant double bonds where
        # chirality is specified (or not). atoms defined by map number
        skip_outcome = False
        for atoms, dirs, is_implicit in reactants.atoms_across_double_bonds:
            if all(i in atoms_rt for i in atoms):
                # All atoms definining chirality were matched to the reactant template
                # So, check if it is consistent with how the template is defined
                #...but /=/ should match \=\ since they are both trans...
                matched_atom_map_nums = tuple(atoms_rt[i].GetAtomMapNum() for i in atoms)

                # Convert atoms_rt to original template's atom map numbers:
                matched_atom_map_nums = tuple(rxn.atoms_rt_idx_to_map[atoms_rt[i].GetIdx()] for i in atoms)

                if matched_atom_map_nums not in rxn.required_rt_bond_defs:
                    continue # this can happen in ring openings, for example
                dirs_template = rxn.required_rt_bond_defs[matched_atom_map_nums]
                if dirs != dirs_template and \
                        (BondDirOpposite[dirs[0]], BondDirOpposite[dirs[1]]) != dirs_template and \
                        not (dirs_template == (BondDir.NONE, BondDir.NONE) and is_implicit):
                    if PLEVEL >= 5: print('Reactant bond chirality does not match template!')
                    if PLEVEL >= 5: print('Based on map numbers...')
                    if PLEVEL >= 5: print('  rct: {} -> {}'.format(matched_atom_map_nums, dirs))
                    if PLEVEL >= 5: print('  tmp: {} -> {}'.format(matched_atom_map_nums, dirs_template))
                    if PLEVEL >= 5: print('skipping this outcome, should not have matched...')
                    skip_outcome = True
                    break
        if skip_outcome:
            continue

        ###############################################################################



        ###############################################################################
        # Convert product(s) to single product so that all
        # reactions can be treated as pseudo-intramolecular
        # But! check for ring openings mistakenly split into multiple
        # This can be diagnosed by duplicate map numbers (i.e., SMILES)

        mapnums = [a.GetAtomMapNum() for m in outcome for a in m.GetAtoms() if a.GetAtomMapNum()]
        if len(mapnums) != len(set(mapnums)): # duplicate?
            if PLEVEL >= 1: print('Found duplicate mapnums in product - need to stitch')
            # need to do a fancy merge
            merged_mol = Chem.RWMol(outcome[0])
            merged_map_to_id = {a.GetAtomMapNum(): a.GetIdx() for a in outcome[0].GetAtoms() if a.GetAtomMapNum()}
            for j in range(1, len(outcome)):
                new_mol = outcome[j]
                for a in new_mol.GetAtoms():
                    if a.GetAtomMapNum() not in merged_map_to_id:
                        merged_map_to_id[a.GetAtomMapNum()] = merged_mol.AddAtom(a)
                for b in new_mol.GetBonds():
                    bi = b.GetBeginAtom().GetAtomMapNum()
                    bj = b.GetEndAtom().GetAtomMapNum()
                    if PLEVEL >= 10: print('stitching bond between {} and {} in stich has chirality {}, {}'.format(
                        bi, bj, b.GetStereo(), b.GetBondDir()
                    ))
                    if not merged_mol.GetBondBetweenAtoms(
                            merged_map_to_id[bi], merged_map_to_id[bj]):
                        merged_mol.AddBond(merged_map_to_id[bi],
                            merged_map_to_id[bj], b.GetBondType())
                        merged_mol.GetBondBetweenAtoms(
                            merged_map_to_id[bi], merged_map_to_id[bj]
                        ).SetStereo(b.GetStereo())
                        merged_mol.GetBondBetweenAtoms(
                            merged_map_to_id[bi], merged_map_to_id[bj]
                        ).SetBondDir(b.GetBondDir())
            outcome = merged_mol.GetMol()
            if PLEVEL >= 1: print('Merged editable mol, converted back to real mol, {}'.format(Chem.MolToSmiles(outcome, True)))
        else:
            new_outcome = outcome[0]
            for j in range(1, len(outcome)):
                new_outcome = AllChem.CombineMols(new_outcome, outcome[j])
            outcome = new_outcome
        if PLEVEL >= 2: print('Converted all outcomes to single molecules')
        ###############################################################################




        ###############################################################################
        # Figure out which atoms were matched in the templates
        # atoms_rt and atoms_p will be outcome-specific.
        atoms_pt = {a.GetAtomMapNum(): atoms_pt_map[a.GetIntProp('old_mapno')] \
            for a in outcome.GetAtoms() if a.HasProp('old_mapno')}
        atoms_p = {a.GetAtomMapNum(): a for a in outcome.GetAtoms() if a.GetAtomMapNum()}

        # Set map numbers of product template
        # note: this is okay to do within the loop, because ALL atoms must be matched
        # in the templates, so the map numbers will get overwritten every time
        # This makes it easier to check parity changes
        [a.SetAtomMapNum(i) for (i, a) in atoms_pt.items()]
        ###############################################################################



        ###############################################################################
        # Check for missing bonds. These are bonds that are present in the reactants,
        # not specified in the reactant template, and not in the product. Accidental
        # fragmentation can occur for intramolecular ring openings
        missing_bonds = []
        for (i, j, b) in reactants.bonds_by_mapnum:
            if i in atoms_p and j in atoms_p:
                # atoms from reactant bond show up in product
                if not outcome.GetBondBetweenAtoms(atoms_p[i].GetIdx(), atoms_p[j].GetIdx()):
                    #...but there is not a bond in the product between those atoms
                    if i not in atoms_rt or j not in atoms_rt or not template_r.GetBondBetweenAtoms(atoms_rt[i].GetIdx(), atoms_rt[j].GetIdx()):
                        # the reactant template did not specify a bond between those atoms (e.g., intentionally destroy)
                        missing_bonds.append((i, j, b))
        if missing_bonds:
            if PLEVEL >= 1: print('Product is missing non-reacted bonds that were present in reactants!')
            outcome = Chem.RWMol(outcome)
            rwmol_map_to_id = {a.GetAtomMapNum(): a.GetIdx() for a in outcome.GetAtoms() if a.GetAtomMapNum()}
            for (i, j, b) in missing_bonds:
                outcome.AddBond(rwmol_map_to_id[i], rwmol_map_to_id[j])
                new_b = outcome.GetBondBetweenAtoms(rwmol_map_to_id[i], rwmol_map_to_id[j])
                new_b.SetBondType(b.GetBondType())
                new_b.SetBondDir(b.GetBondDir())
                new_b.SetIsAromatic(b.GetIsAromatic())
            outcome = outcome.GetMol()
            atoms_p = {a.GetAtomMapNum(): a for a in outcome.GetAtoms() if a.GetAtomMapNum()}
        else:
            if PLEVEL >= 3: print('No missing bonds')
        ###############################################################################


        # Now that we've fixed any bonds, connectivity is set. This is a good time
        # to udpate the property cache, since all that is left is fixing atom/bond
        # stereochemistry.
        try:
            Chem.SanitizeMol(outcome)
            outcome.UpdatePropertyCache()
        except ValueError as e:
            if PLEVEL >= 1: print('{}, {}'.format(Chem.MolToSmiles(outcome, True), e))
            continue


        ###############################################################################
        # Correct tetra chirality in the outcome
        tetra_copied_from_reactants = []
        for a in outcome.GetAtoms():
            # Participants in reaction core (from reactants) will have old_mapno
            # Spectators present in reactants will have react_atom_idx
            # ...so new atoms will have neither!
            if not a.HasProp('old_mapno'):
                # Not part of the reactants template

                if not a.HasProp('react_atom_idx'):
                    # Atoms only appear in product template - their chirality
                    # should be properly instantiated by RDKit...hopefully...
                    if PLEVEL >= 4: print('Atom {} created by product template, should have right chirality'.format(a.GetAtomMapNum()))

                else:
                    if PLEVEL >= 4: print('Atom {} outside of template, copy chirality from reactants'.format(a.GetAtomMapNum()))
                    copy_chirality(atoms_r[a.GetAtomMapNum()], a)
                    if a.GetChiralTag() != ChiralType.CHI_UNSPECIFIED:
                        tetra_copied_from_reactants.append(a)

            else:
                # Part of reactants and reaction core

                if template_atom_could_have_been_tetra(atoms_rt[a.GetAtomMapNum()]):
                    if PLEVEL >= 3: print('Atom {} was in rct template (could have been tetra)'.format(a.GetAtomMapNum()))

                    if template_atom_could_have_been_tetra(atoms_pt[a.GetAtomMapNum()]):
                        if PLEVEL >= 3: print('Atom {} in product template could have been tetra, too'.format(a.GetAtomMapNum()))

                        # Was the product template specified?

                        if atoms_pt[a.GetAtomMapNum()].GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
                            # No, leave unspecified in product
                            if PLEVEL >= 3: print('...but it is not specified in product, so destroy chirality')
                            a.SetChiralTag(ChiralType.CHI_UNSPECIFIED)

                        else:
                            # Yes
                            if PLEVEL >= 3: print('...and product is specified')

                            # Was the reactant template specified?

                            if atoms_rt[a.GetAtomMapNum()].GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
                                # No, so the reaction introduced chirality
                                if PLEVEL >= 3: print('...but reactant template was not, so copy from product template')
                                copy_chirality(atoms_pt[a.GetAtomMapNum()], a)

                            else:
                                # Yes, so we need to check if chirality should be preserved or inverted
                                if PLEVEL >= 3: print('...and reactant template was, too! copy from reactants')
                                copy_chirality(atoms_r[a.GetAtomMapNum()], a)
                                if atom_chirality_matches(atoms_pt[a.GetAtomMapNum()], atoms_rt[a.GetAtomMapNum()]) == -1:
                                    if PLEVEL >= 3: print('but! reactant template and product template have opposite stereochem, so invert')
                                    a.InvertChirality()

                    else:
                        # Reactant template chiral, product template not - the
                        # reaction is supposed to destroy chirality, so leave
                        # unspecified
                        if PLEVEL >= 3: print('If reactant template could have been ' +
                            'chiral, but the product template could not, then we dont need ' +
                            'to worry about specifying product atom chirality')

                else:
                    if PLEVEL >= 3: print('Atom {} could not have been chiral in reactant template'.format(a.GetAtomMapNum()))

                    if not template_atom_could_have_been_tetra(atoms_pt[a.GetAtomMapNum()]):
                        if PLEVEL >= 3: print('Atom {} also could not have been chiral in product template', a.GetAtomMapNum())
                        if PLEVEL >= 3: print('...so, copy chirality from reactant instead')
                        copy_chirality(atoms_r[a.GetAtomMapNum()], a)
                        if a.GetChiralTag() != ChiralType.CHI_UNSPECIFIED:
                            tetra_copied_from_reactants.append(a)

                    else:
                        if PLEVEL >= 3: print('Atom could/does have product template chirality!'.format(a.GetAtomMapNum()))
                        if PLEVEL >= 3: print('...so, copy chirality from product template')
                        copy_chirality(atoms_pt[a.GetAtomMapNum()], a)

            if PLEVEL >= 3: print('New chiral tag {}'.format(a.GetChiralTag()))
        if skip_outcome:
            if PLEVEL >= 2: print('Skipping this outcome - chirality broken?')
            continue
        if PLEVEL >= 2: print('After attempting to re-introduce chirality, outcome = {}'.format(Chem.MolToSmiles(outcome, True)))
        ###############################################################################


        ###############################################################################
        # Correct bond directionality in the outcome
        for b in outcome.GetBonds():
            if b.GetBondType() != BondType.DOUBLE:
                continue

            # Ring double bonds do not need to be touched(?)
            if b.IsInRing():
                continue

            ba = b.GetBeginAtom()
            bb = b.GetEndAtom()

            # Is it possible at all to specify this bond?
            if ba.GetDegree() == 1 or bb.GetDegree() == 1:
                continue

            if PLEVEL >= 5: print('Looking at outcome bond {}={}'.format(ba.GetAtomMapNum(), bb.GetAtomMapNum()))

            if ba.HasProp('old_mapno') and bb.HasProp('old_mapno'):
                # Need to rely on templates for bond chirality, both atoms were
                # in the reactant template
                if PLEVEL >= 5: print('Both atoms in this double bond were in the reactant template')
                if (ba.GetIntProp('old_mapno'), bb.GetIntProp('old_mapno')) in \
                        rxn.required_bond_defs_coreatoms:
                    if PLEVEL >= 5: print('and reactant template *could* have specified the chirality!')
                    if PLEVEL >= 5: print('..product should be property instantiated')
                    continue
                if PLEVEL >= 5: print('But it was impossible to have specified chirality (e.g., aux C=C for context)')

            elif not ba.HasProp('react_atom_idx') and not bb.HasProp('react_atom_idx'):
                # The atoms were both created by the product template, so any bond
                # stereochemistry should have been instantiated by the product template
                # already...hopefully...otherwise it isn't specific enough?
                continue

            # Need to copy from reactants, this double bond was simply carried over,
            # *although* one of the atoms could have reacted and been an auxilliary
            # atom in the reaction, e.g., C/C=C(/CO)>>C/C=C(/C[Br])
            if PLEVEL >= 5: print('Restoring cis/trans character of bond {}={} from reactants'.format(
                ba.GetAtomMapNum(), bb.GetAtomMapNum()))

            # Start with setting the BeginAtom
            begin_atom_specified = restore_bond_stereo_to_sp2_atom(ba, reactants.bond_dirs_by_mapnum)

            if not begin_atom_specified:
                # don't bother setting other side of bond, since we won't be able to
                # fully specify this bond as cis/trans
                continue

            # Look at other side of the bond now, the EndAtom
            end_atom_specified = restore_bond_stereo_to_sp2_atom(bb, reactants.bond_dirs_by_mapnum)
            if not end_atom_specified:
                # note: this can happen if C=C/C-N turns into C=C/C=N
                if PLEVEL >= 1:
                    print(reactants.bond_dirs_by_mapnum)
                    print(ba.GetAtomMapNum())
                    print(bb.GetAtomMapNum())
                    print(Chem.MolToSmiles(reactants.reactants, True))
                    print(Chem.MolToSmiles(outcome, True))
                    print('Uh oh, looks like bond direction is only specified for half of this bond?')

        ###############################################################################

        #Keep track of the reacting atoms for later use in grouping
        atoms_diff = {x:atoms_are_different(atoms_r[x],atoms_p[x]) for x in atoms_rt}
        #make tuple of changed atoms
        atoms_changed = tuple([x for x in atoms_diff.keys() if atoms_diff[x] == True])
        mapped_outcome = Chem.MolToSmiles(outcome, True)

        if not keep_mapnums:
            for a in outcome.GetAtoms():
                a.SetAtomMapNum(0)

        # Now, check to see if we have destroyed chirality
        # this occurs when chirality was not actually possible (e.g., due to
        # symmetry) but we had assigned a tetrahedral center originating
        # from the reactants.
        #    ex: SMILES C(=O)1C[C@H](Cl)CCC1
        #        SMARTS [C:1]-[C;H0;D3;+0:2](-[C:3])=[O;H0;D1;+0]>>[C:1]-[CH2;D2;+0:2]-[C:3]
        skip_outcome = False
        if len(tetra_copied_from_reactants) > 0:
            Chem.AssignStereochemistry(outcome, cleanIt=True, force=True)
            for a in tetra_copied_from_reactants:
                if a.GetChiralTag() == ChiralType.CHI_UNSPECIFIED:
                    if PLEVEL >= 2: print('Auxiliary reactant atom was chiral, now is broken -> skip outcome')
                    skip_outcome = True
                    break
        if skip_outcome:
            continue


        smiles = Chem.MolToSmiles(outcome, True)
        smiles_new = canonicalize_outcome_smiles(smiles)
        if smiles_new is None:
            continue

        final_outcomes.add(smiles_new)
        mapped_outcomes[smiles_new] = (mapped_outcome, atoms_changed)
    ###############################################################################
    # One last fix for consolidating multiple stereospecified products...
    if combine_enantiomers:
        final_outcomes = combine_enantiomers_into_racemic(final_outcomes)
    ###############################################################################
    if return_mapped:
        return list(final_outcomes), mapped_outcomes
    else:
        return list(final_outcomes)

if __name__ == '__main__':
    # Directly use SMILES/SMARTS
    reaction_smarts = '[C:1][OH:2]>>[C:1][O:2][C]'
    reactant_smiles = 'OCC(=O)OCCCO'
    outcomes = rdchiralRunText(reaction_smarts, reactant_smiles)
    print(outcomes)

    reaction_smarts = '[CH3;D1;+0:1]-[C:2]=[C:3]-[#7:4]-[C:5](=[O;D1;H0:6])-[C:7]-[NH2;D1;+0:8]>>C-C(=O)-O-[CH2;D2;+0:1]-[C:2]=[C:3]-[#7:4]-[C:5](=[O;D1;H0:6])-[C:7]-[NH;D2;+0:8]-C(=O)-C(-O-C(-C)=O)-c1:c:c:c:c:c:1'
    reactant_smiles = '[CH3:5][C:6]1[CH2:28][S:27][C@@H:9]2[C@H:10]([NH2:13])[C:11](=[O:12])[N:8]2[C:7]=1[C:29]([OH:31])=[O:30]'
    outcomes = rdchiralRunText(reaction_smarts, reactant_smiles)
    print(outcomes)

    # Pre-initialize
    rxn = rdchiralReaction(reaction_smarts)
    reactants = rdchiralReactants(reactant_smiles)
    outcomes = rdchiralRun(rxn, reactants)
    print(outcomes)

    # Get list of atoms that changed as well
    outcomes, mapped_outcomes = rdchiralRun(rxn, reactants, return_mapped=True)
    print(outcomes, mapped_outcomes)



import math
import torch
import torch.nn as nn
import torch.nn.functional as F


class SSP(nn.Softplus):
    def __init__(self, beta=1, threshold=20):
        super(SSP, self).__init__(beta, threshold)

    def forward(self, input):
        sp0 = F.softplus(torch.zeros(1), self.beta, self.threshold).item()
        return F.softplus(input, self.beta, self.threshold) - sp0


class PositionwiseFeedForward(nn.Module):
    """ A two-layer Feed-Forward-Network with residual layer norm.

        Args:
            d_model (int): the size of input for the first-layer of the FFN.
            d_ff (int): the hidden layer size of the second-layer
                              of the FNN.
            dropout (float): dropout probability(0-1.0).
    """

    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.layer_norm = LayerNorm(d_model)
        self.dropout_1 = nn.Dropout(dropout)
        self.relu = nn.ReLU()
        self.dropout_2 = nn.Dropout(dropout)

    def forward(self, x):
        """
        Layer definition.

        Args:
            input: [ batch_size, input_len, model_dim ]


        Returns:
            output: [ batch_size, input_len, model_dim ]
        """
        inter = self.dropout_1(self.relu(self.w_1(self.layer_norm(x))))
        output = self.dropout_2(self.w_2(inter))
        return output + x


class MultiHeadedAttention(nn.Module):
    """
    Multi-Head Attention module from OpenNMT
    """

    def __init__(self, head_count, model_dim, dropout=0.1):
        assert model_dim % head_count == 0
        self.dim_per_head = model_dim // head_count
        self.model_dim = model_dim

        super(MultiHeadedAttention, self).__init__()
        self.head_count = head_count
        self.linear_keys = nn.Linear(model_dim,
                                     head_count * self.dim_per_head)
        self.linear_values = nn.Linear(model_dim,
                                       head_count * self.dim_per_head)
        self.linear_query = nn.Linear(model_dim,
                                      head_count * self.dim_per_head)
        self.softmax = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.final_linear = nn.Linear(model_dim, model_dim)

        self.edge_project = nn.Sequential(nn.Linear(model_dim, model_dim),
                                          SSP(),
                                          nn.Linear(model_dim, model_dim // 2))
        self.edge_update = nn.Sequential(nn.Linear(model_dim * 2, model_dim),
                                         SSP(),
                                         nn.Linear(model_dim, model_dim))

    def forward(self, key, value, query, mask=None, additional_mask=None,
                layer_cache=None, type=None, edge_feature=None, pair_indices=None):

        """
        Compute the context vector and the attention vectors.

        Args:
           key (`FloatTensor`): set of `key_len`
                key vectors `[batch, key_len, dim]`
           value (`FloatTensor`): set of `key_len`
                value vectors `[batch, key_len, dim]`
           query (`FloatTensor`): set of `query_len`
                 query vectors  `[batch, query_len, dim]`
           mask: binary mask indicating which keys have
                 non-zero attention `[batch, query_len, key_len]`
        Returns:
           (`FloatTensor`, `FloatTensor`) :

           * output context vectors `[batch, query_len, dim]`
           * one of the attention vectors `[batch, query_len, key_len]`
        """
        global query_projected, key_shaped, value_shaped
        batch_size = key.size(0)
        dim_per_head = self.dim_per_head
        head_count = self.head_count
        key_len = key.size(1)
        query_len = query.size(1)

        def shape(x):
            """  projection """
            return x.view(batch_size, -1, head_count, dim_per_head) \
                .transpose(1, 2)

        def unshape(x):
            """  compute context """
            return x.transpose(1, 2).contiguous() \
                .view(batch_size, -1, head_count * dim_per_head)

        # 1) Project key, value, and query.
        if layer_cache is not None:
            if type == "self":
                query_projected, key_projected, value_projected = self.linear_query(query), \
                                                                  self.linear_keys(query), \
                                                                  self.linear_values(query)

                key_shaped = shape(key_projected)
                value_shaped = shape(value_projected)

                if layer_cache is not None:
                    device = key.device
                    if layer_cache["self_keys"] is not None:
                        key_shaped = torch.cat(
                            (layer_cache["self_keys"].to(device), key_shaped),
                            dim=2)
                    if layer_cache["self_values"] is not None:
                        value_shaped = torch.cat(
                            (layer_cache["self_values"].to(device), value_shaped),
                            dim=2)
                    layer_cache["self_keys"] = key_shaped
                    layer_cache["self_values"] = value_shaped
            elif type == "context":
                query_projected = self.linear_query(query)
                if layer_cache is not None:
                    if layer_cache["memory_keys"] is None:
                        key_projected, value_projected = self.linear_keys(key), \
                                                         self.linear_values(value)
                        key_shaped = shape(key_projected)
                        value_shaped = shape(value_projected)
                    else:
                        key_shaped, value_shaped = layer_cache["memory_keys"], \
                                                   layer_cache["memory_values"]
                    layer_cache["memory_keys"] = key_shaped
                    layer_cache["memory_values"] = value_shaped
                else:
                    key_projected, value_projected = self.linear_keys(key), \
                                                     self.linear_values(value)
                    key_shaped = shape(key_projected)
                    value_shaped = shape(value_projected)
        else:
            key_projected = self.linear_keys(key)
            value_projected = self.linear_values(value)
            query_projected = self.linear_query(query)
            key_shaped = shape(key_projected)
            value_shaped = shape(value_projected)

        query_shaped = shape(query_projected)
        key_len = key_shaped.size(2)
        query_len = query_shaped.size(2)

        # 2) Calculate and scale scores (local and global attention heads).
        # 3) Apply attention dropout and compute context vectors.
        # Local-Global Decoder:
        if edge_feature is None and additional_mask is not None:
            query_shaped = query_shaped / math.sqrt(dim_per_head)
            query_shaped_global, query_shaped_local = query_shaped[:, :head_count // 2], query_shaped[:,
                                                                                         head_count // 2:]
            key_shaped_global, key_shaped_local = key_shaped[:, :head_count // 2], key_shaped[:, head_count // 2:]
            value_shaped_global, value_shaped_local = value_shaped[:, :head_count // 2], value_shaped[:,
                                                                                         head_count // 2:]

            # Global:
            score_global = torch.matmul(query_shaped_global, key_shaped_global.transpose(2, 3))
            top_score = score_global.view(batch_size, score_global.shape[1],
                                          query_len, key_len)[:, 0, :, :].contiguous()
            if mask is not None:
                mask = mask.unsqueeze(1).expand_as(score_global).clone()
                score_global = score_global.masked_fill(mask, -1e18)
            attn = self.softmax(score_global)
            drop_attn = self.dropout(attn)
            global_context = torch.matmul(drop_attn, value_shaped_global)

            # Local:
            score_local = torch.matmul(query_shaped_local, key_shaped_local.transpose(2, 3))
            if additional_mask is not None:
                additional_mask = additional_mask.unsqueeze(1).unsqueeze(2).expand_as(score_local).clone()
                score_local = score_local.masked_fill(additional_mask, -1e18)
            attn = self.softmax(score_local)
            drop_attn = self.dropout(attn)
            local_context = torch.matmul(drop_attn, value_shaped_local)

            # Merge:
            context = torch.cat([global_context, local_context], dim=1)
            context = unshape(context)

        # Local-Global Encoder:
        elif edge_feature is not None:
            # Local: (node update)
            edge_feature_shaped = self.edge_project(edge_feature).view(-1, head_count // 2, dim_per_head)
            key_shaped_local = key_shaped[pair_indices[0], head_count // 2:, pair_indices[2]]
            query_shaped_local = query_shaped[pair_indices[0], head_count // 2:, pair_indices[1]]
            value_shaped_local = value_shaped[:, head_count // 2:]

            key_shaped_local = key_shaped_local * edge_feature_shaped
            query_shaped_local = query_shaped_local / math.sqrt(dim_per_head)

            scores_local = torch.matmul(query_shaped_local.unsqueeze(2),
                                        key_shaped_local.unsqueeze(3)).view(edge_feature.shape[0], head_count // 2)

            score_expand_local = scores_local.new_full(
                (value.shape[0], value.shape[1], value.shape[1], head_count // 2), -float('inf'))
            score_expand_local[pair_indices] = scores_local
            score_expand_local = score_expand_local.transpose(1, 3).transpose(2, 3)

            attn_local = self.softmax(score_expand_local)
            attn_local = attn_local.masked_fill(score_expand_local < -10000, 0)
            drop_attn_local = self.dropout(attn_local)
            local_context = torch.matmul(drop_attn_local, value_shaped_local)

            # Global：
            query_shaped_global = query_shaped[:, :head_count // 2]
            key_shaped_global = key_shaped[:, :head_count // 2]
            value_shaped_global = value_shaped[:, :head_count // 2]

            query_shaped_global = query_shaped_global / math.sqrt(dim_per_head)
            score_global = torch.matmul(query_shaped_global, key_shaped_global.transpose(2, 3))
            top_score = score_global.view(batch_size, score_global.shape[1],
                                          query_len, key_len)[:, 0, :, :].contiguous()
            if mask is not None:
                mask = mask.unsqueeze(1).expand_as(score_global).clone()
                score_global = score_global.masked_fill(mask, -1e18)

            attn = self.softmax(score_global)
            drop_attn = self.dropout(attn)
            global_context = torch.matmul(drop_attn, value_shaped_global)

            # Merge:
            context = torch.cat([global_context, local_context], dim=1)
            context = unshape(context)

        # Original Encoder/Decoder/Decoder self-attention:
        else:
            query_shaped = query_shaped / math.sqrt(dim_per_head)
            scores = torch.matmul(query_shaped, key_shaped.transpose(2, 3))
            top_score = scores.view(batch_size, scores.shape[1],
                                    query_len, key_len)[:, 0, :, :].contiguous()
            if mask is not None:
                mask = mask.unsqueeze(1).expand_as(scores).clone()
                if additional_mask is not None:  # Local head of decoder:
                    additional_mask = additional_mask.unsqueeze(1).expand((batch_size, head_count // 2,
                                                                           query_len, key_len))
                    mask[:, mask.shape[1] // 2:] = additional_mask
                scores = scores.masked_fill(mask, -1e18)
            attn = self.softmax(scores)
            drop_attn = self.dropout(attn)
            context = torch.matmul(drop_attn, value_shaped)
            context = unshape(context)

        output = self.final_linear(context)

        # 4) Edge update (only in encoder)
        if edge_feature is not None:
            node_feature_updated = output
            node_features = torch.cat(
                [node_feature_updated[pair_indices[0], pair_indices[1]],
                 node_feature_updated[pair_indices[0], pair_indices[2]]], dim=-1)
            edge_feature_updated = self.edge_update(node_features)

            return output, top_score, edge_feature_updated
        else:
            return output, top_score, None


class LayerNorm(nn.Module):
    """
        Layer Normalization class
    """

    def __init__(self, features, eps=1e-6):
        super(LayerNorm, self).__init__()
        self.a_2 = nn.Parameter(torch.ones(features))
        self.b_2 = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2
